#!/usr/bin/python
# -*- coding: utf-8 -*-
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
from sklearn.metrics import log_loss

from SMdRQA.RQA_functions import doanes_formula
from SMdRQA.RQA_functions import binscalc
from SMdRQA.RQA_functions import mutualinfo
from SMdRQA.RQA_functions import timedelayMI
from SMdRQA.RQA_functions import findtau
from SMdRQA.RQA_functions import delayseries
from SMdRQA.RQA_functions import nearest
from SMdRQA.RQA_functions import fnnratio
from SMdRQA.RQA_functions import fnnhitszero
from SMdRQA.RQA_functions import findm
from SMdRQA.RQA_functions import reccplot
from SMdRQA.RQA_functions import reccrate
from SMdRQA.RQA_functions import findeps
from SMdRQA.RQA_functions import plotwindow
from SMdRQA.RQA_functions import vert_hist
from SMdRQA.RQA_functions import onedhist
from SMdRQA.RQA_functions import diaghist
from SMdRQA.RQA_functions import percentmorethan
from SMdRQA.RQA_functions import mode
from SMdRQA.RQA_functions import maxi
from SMdRQA.RQA_functions import average
from SMdRQA.RQA_functions import entropy

import os
import pickle
import pandas as pd
import seaborn as sns
from random import sample
import scipy.stats as ss
from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_selection import SelectFromModel
from sklearn.feature_selection import RFE
from sklearn.metrics import auc
from sklearn.metrics import accuracy_score
from sklearn.metrics import log_loss
from sklearn.preprocessing import normalize
import matplotlib.pyplot as plt
import itertools
from tqdm import tqdm


def HistogramSampling(histogram, bins_midpoints, samples):
    '''\n    Function for doing histogram sampling from a distribution\n\n    Parameters\n    ----------\n\n    histogram : array\n        histogram counts normalised to get a sum =1\n\n      bins_midpoints : array\n          values corresponding to the probability values given as "histogram"\n\n      samples : int\n          number of histograms samples\n\n      Returns\n      -------\n\n      hist2 : array\n           histogram counts of the resulting distribution\n\n      bins_midpoints : array\n           midpoints of bins corresponding to hist2\n\n      References\n      ----------\n      - Marwan, N., Romano, M. C., Thiel, M., & Kurths, J. (2007). Recurrence plots for the analysis of complex systems. Physics reports, 438 (5-6), 237\xe2\x80\x93329.\n\n      '''

    cdf = np.cumsum(histogram)
    cdf = cdf / cdf[-1]
    values = np.random.rand(samples)
    value_bins = np.searchsorted(cdf, values)

    # print('values:',values)
    # print('value bins:',value_bins)

    random_from_cdf = bins_midpoints[value_bins]
    edges = list(bins_midpoints - 0.5)
    edges.append(bins_midpoints[-1] + 0.5)

    (hist2, edges1) = np.histogram(random_from_cdf, bins=edges)

    # print('bin_mids:',(edges1[:-1]+edges1[1:])/2)

    return (random_from_cdf, hist2)


# for percenthage laminarity

def Sliding_windowBS_sub_per_lam(
    RP,
    maxsize,
    winsize,
    n_boot,
):
    '''\n    Function for computing the CI of the bootstrapped confidence interval for percentage laminarity\n\n    Parameters\n    ----------\n\n    RP : ndarray\n        recurrence plot\n\n    maxsize : int\n        RP size\n\n    winsize : int\n        size of the sliding window\n\n     n_boot : int\n        number of bootstrap samples\n\n    Returns\n    -------\n\n    CI : double\n        confidence interval(95% quantile - 5% quantile)\n\n\n    References\n    ----------\n    - Marwan, N., Romano, M. C., Thiel, M., & Kurths, J. (2007). Recurrence plots for the analysis of complex systems. Physics reports, 438 (5-6), 237\xe2\x80\x93329.\n\n    '''

    num_windows = maxsize - winsize + 1
    P = []
    P_bar = []
    for i in range(num_windows):
        sub_window = RP[i:i + winsize, i:i + winsize]
        diag_his = vert_hist(sub_window, winsize)

        # diag_his=diag_his[1:]

        P.append(diag_his / np.sum(diag_his))
        P_bar.append(diag_his)
    P = np.array(P)
    n_bar = int(np.sum(P_bar) / num_windows)
    print('n_bar:', n_bar)

    # print(np.sum(P, axis=1))

    P_summed = np.sum(P, axis=0)
    P_summed = P_summed / np.sum(P_summed)

    # print('n_bar:',n_bar)

    bins_midpoints = np.array(range(1, len(P_summed) + 1))

    # line_array=IND_array+1

    VAR = []
    for j in range(n_boot):
        (hist_sample, hist) = HistogramSampling(P_summed,
                                                bins_midpoints, n_bar)
        hist = hist / np.sum(hist)

        # percent_more=entropy(hist,2, winsize)

        percent_more = np.sum(hist[1:]) / np.sum(hist)

        # avg_line=entropy(sel_P,1,len(sel_P)-1)
        # print(avg_line)
        # print(avg_line)

        VAR.append(percent_more)

    # return np.std(VAR)**2

    return np.quantile(VAR, 0.95) - np.quantile(VAR, 0.05)


def Sliding_windowBS_sub_avg_vert(
    RP,
    maxsize,
    winsize,
    n_boot,
):
    '''\n    Function for computing the CI of the bootstrapped confidence interval for average vertical line length\n\n    Parameters\n    ----------\n\n    RP : ndarray\n        recurrence plot\n\n    maxsize : int\n        RP size\n\n    winsize : int\n        size of the sliding window\n\n     n_boot : int\n        number of bootstrap samples\n\n    Returns\n    -------\n\n    CI : double\n        confidence interval(95% quantile - 5% quantile)\n\n\n    References\n    ----------\n    - Marwan, N., Romano, M. C., Thiel, M., & Kurths, J. (2007). Recurrence plots for the analysis of complex systems. Physics reports, 438 (5-6), 237\xe2\x80\x93329.\n\n    '''

    num_windows = maxsize - winsize + 1
    P = []
    P_bar = []
    for i in range(num_windows):
        sub_window = RP[i:i + winsize, i:i + winsize]
        diag_his = vert_hist(sub_window, winsize)

        # diag_his=diag_his[1:]

        P.append(diag_his / np.sum(diag_his))
        P_bar.append(diag_his)
    P = np.array(P)
    n_bar = int(np.sum(np.mean(P_bar, axis=0)))
    print('n_bar:', n_bar)

    # print(np.sum(P, axis=1))

    P_summed = np.sum(P, axis=0)
    P_summed = P_summed / np.sum(P_summed)

    # print('n_bar:',n_bar)

    bins_midpoints = np.array(range(1, len(P_summed) + 1))

    # line_array=IND_array+1

    VAR = []
    for j in range(n_boot):
        (hist_sample, hist) = HistogramSampling(P_summed,
                                                bins_midpoints, n_bar)
        hist = hist / np.sum(hist)

        # percent_more=entropy(hist,2, winsize)

        percent_more = np.mean(hist_sample[1:])

        # avg_line=entropy(sel_P,1,len(sel_P)-1)
        # print(avg_line)
        # print(avg_line)

        VAR.append(percent_more)

    # return np.std(VAR)**2

    return np.quantile(VAR, 0.95) - np.quantile(VAR, 0.05)


def Sliding_windowBS_sub_percent_det(
    RP,
    maxsize,
    winsize,
    n_boot,
):
    '''\n    Function for computing the CI of the bootstrapped confidence interval for percentage determinism\n\n    Parameters\n    ----------\n\n    RP : ndarray\n        recurrence plot\n\n    maxsize : int\n        RP size\n\n    winsize : int\n        size of the sliding window\n\n     n_boot : int\n        number of bootstrap samples\n\n    Returns\n    -------\n\n    CI : double\n        confidence interval(95% quantile - 5% quantile)\n\n\n    References\n    ----------\n    - Marwan, N., Romano, M. C., Thiel, M., & Kurths, J. (2007). Recurrence plots for the analysis of complex systems. Physics reports, 438 (5-6), 237\xe2\x80\x93329.\n\n    '''

    num_windows = maxsize - winsize + 1
    P = []
    P_bar = []
    for i in range(num_windows):
        sub_window = RP[i:i + winsize, i:i + winsize]
        diag_his = diaghist(sub_window, winsize)

        # diag_his=diag_his[1:]

        P.append(diag_his / np.sum(diag_his))
        P_bar.append(diag_his)
    P = np.array(P)
    n_bar = int(np.sum(P_bar) / num_windows)
    print('n_bar:', n_bar)

    # print(np.sum(P, axis=1))

    P_summed = np.sum(P, axis=0)
    P_summed = P_summed / np.sum(P_summed)

    # print('n_bar:',n_bar)

    bins_midpoints = np.array(range(1, len(P_summed) + 1))

    # line_array=IND_array+1

    VAR = []
    for j in range(n_boot):
        (hist_sample, hist) = HistogramSampling(P_summed,
                                                bins_midpoints, n_bar)
        hist = hist / np.sum(hist)

        # percent_more=entropy(hist,2, winsize)

        percent_more = np.sum(hist[1:]) / np.sum(hist)

        # avg_line=entropy(sel_P,1,len(sel_P)-1)
        # print(avg_line)
        # print(avg_line)

        VAR.append(percent_more)

    # return np.std(VAR)**2

    return np.quantile(VAR, 0.95) - np.quantile(VAR, 0.05)


def Sliding_windowBS_sub_avg_diag(
    RP,
    maxsize,
    winsize,
    n_boot,
):
    '''\n    Function for computing the CI of the bootstrapped confidence interval for average diagonal line length\n\n    Parameters\n    ----------\n\n    RP : ndarray\n        recurrence plot\n\n    maxsize : int\n        RP size\n\n    winsize : int\n        size of the sliding window\n\n     n_boot : int\n        number of bootstrap samples\n\n    Returns\n    -------\n\n    CI : double\n        confidence interval(95% quantile - 5% quantile)\n\n\n    References\n    ----------\n    - Marwan, N., Romano, M. C., Thiel, M., & Kurths, J. (2007). Recurrence plots for the analysis of complex systems. Physics reports, 438 (5-6), 237\xe2\x80\x93329.\n\n    '''

    num_windows = maxsize - winsize + 1
    P = []
    P_bar = []
    for i in range(num_windows):
        sub_window = RP[i:i + winsize, i:i + winsize]
        diag_his = diaghist(sub_window, winsize)

        # diag_his=diag_his[1:]

        P.append(diag_his / np.sum(diag_his))
        P_bar.append(diag_his)
    P = np.array(P)
    n_bar = int(np.sum(np.mean(P_bar, axis=0)))
    print('n_bar:', n_bar)

    # print(np.sum(P, axis=1))

    P_summed = np.sum(P, axis=0)
    P_summed = P_summed / np.sum(P_summed)

    # print('n_bar:',n_bar)

    bins_midpoints = np.array(range(1, len(P_summed) + 1))

    # line_array=IND_array+1

    VAR = []
    for j in range(n_boot):
        (hist_sample, hist) = HistogramSampling(P_summed,
                                                bins_midpoints, n_bar)
        hist = hist / np.sum(hist)

        # percent_more=entropy(hist,2, winsize)

        percent_more = np.mean(hist_sample[1:])

        # avg_line=entropy(sel_P,1,len(sel_P)-1)
        # print(avg_line)
        # print(avg_line)

        VAR.append(percent_more)

    # return np.std(VAR)**2

    return np.quantile(VAR, 0.95) - np.quantile(VAR, 0.05)


def Sliding_windowBS(
    RP,
    maxsize,
    var,
    n_boot=1000,
):
    '''\n    Function for computing the CI of the bootstrapped confidence interval for a given variable for different window sizes starting from 20 to a maximum of RP size incremented by value of 10\n\n    Parameters\n    ----------\n\n    RP : ndarray\n        recurrence plot\n\n    maxsize : int\n        RP size\n\n    var : str\n        RQA variable name\n\n        *percent_lam* : percentage laminarity\n\n        *percent_det* : percentage determinism\n\n        *avg_vert* : average value of vertical line distribution\n\n        *avg_diag* : average value of diagonal line distribution\n\n     n_boot : int\n        number of bootstrap samples\n\n    Returns\n    -------\n\n    data : dataframe\n        dataframe containing the CI estimate of corresponding window sizes, columns are\n\n        *WINSIZE* : window size\n\n        *95% quantile- 5% quantile* : confidence interval\n\n\n    References\n    ----------\n    - Marwan, N., Romano, M. C., Thiel, M., & Kurths, J. (2007). Recurrence plots for the analysis of complex systems. Physics reports, 438 (5-6), 237\xe2\x80\x93329.\n\n    '''

    WINSIZE = []
    VAR = []
    for i in tqdm(range(20, maxsize, 10)):
        if var == 'percent_lam':
            var_std = Sliding_windowBS_sub_per_lam(RP, maxsize, i,
                                                   n_boot)
        elif var == 'percent_det':
            var_std = Sliding_windowBS_sub_percent_det(RP, maxsize, i,
                                                       n_boot)
        elif var == 'avg_vert':
            var_std = Sliding_windowBS_sub_avg_vert(RP, maxsize, i,
                                                    n_boot)
        elif var == 'avg_diag':
            var_std = Sliding_windowBS_sub_avg_diag(RP, maxsize, i,
                                                    n_boot)

        WINSIZE.append(i)
        VAR.append(var_std)

    DICT = {'WINSIZE': WINSIZE, '95% quantile- 5% quantile': VAR}
    df = pd.DataFrame.from_dict(DICT)
    return df


def Sliding_window_whole_data(RP_dir, var, n_boot=1000):
    '''\n    Function for computing the CI of the bootstrapped confidence interval for a given variable for different window sizes starting from 20 to a maximum of RP size incremented by value of 10. Done for all datasets in the folder specified\n\n    Parameters\n    ----------\n\n    RP_dir : str\n        directory containing all the RP files(.npy)\n\n    var : str\n        RQA variable name\n\n        *percent_lam* : percentage laminarity\n\n        *percent_det* : percentage determinism\n\n        *avg_vert* : average value of vertical line distribution\n\n        *avg_diag* : average value of diagonal line distribution\n\n     n_boot : int\n        number of bootstrap samples\n\n    Returns\n    -------\n\n    data : dataframe\n        dataframe containing the CI estimate of corresponding window sizes, columns are\n\n        *WINSIZE* : window size\n\n        *95% quantile- 5% quantile* : confidence interval\n\n        *group* : filename of the RPs\n\n\n    References\n    ----------\n    - Marwan, N., Romano, M. C., Thiel, M., & Kurths, J. (2007). Recurrence plots for the analysis of complex systems. Physics reports, 438 (5-6), 237\xe2\x80\x93329.\n\n    '''

    files = os.listdir(RP_dir)
    DFs = []
    for File in files:
        path = RP_dir + '/' + File
        info = File.split('-')[0]
        RP = np.load(path)
        maxsize = len(RP)
        df = Sliding_windowBS(RP, maxsize, var, n_boot=n_boot)
        df['group'] = info
        DFs.append(df)

    df_out = pd.concat(DFs)

    return df_out



import pandas as pd
import numpy as np
from itertools import combinations
from sklearn.model_selection import RepeatedStratifiedKFold
from sklearn.neighbors import KNeighborsClassifier
from sklearn.metrics import accuracy_score, roc_auc_score
from sklearn.preprocessing import StandardScaler


def feature_selection(x_train, y_train, features, y_name, inner_splits, select_k=False, repeats=5):  # inner loop function
    '''
    Function used for feature selection  using best subset setection and average cross validation accuracy

    Parameters
    ----------

    x_train : array
        training set from the outer loops of nested cross validation function

    y_train : array
        output variables for the training set from the outer loops of the nested cross validation

    features : list
        total set of features from which features are being selected

    y_name : str
        column name for the outcome variable

    inner_splits : int
        cross validation splits for feature selection

    select_k : bool
        whether the user wants to tune k also based on training set, default = False, k=5

    repeats : int
        number of repeats for cross validation for feature selection

    Returns
    -------

    best_features : list
        list of features got selected

    best_score : double
         average accuracy for the best feature subset

    best_roc_auc : double
         average ROC AUC for the best feature subset

    References
    ----------
    - `Scikit-Learn Nested Cross-Validation Example <https://scikit-learn.org/stable/auto_examples/model_selection/plot_nested_cross_validation_iris.html>`__

    - Cawley, Gavin C., and Nicola LC Talbot. "On over-fitting in model selection and subsequent selection bias in performance evaluation." The Journal of Machine Learning Research 11 (2010): 2079-2107.

    '''
    from tqdm import tqdm
    # Load the data into a pandas DataFrame

    if not select_k:
        X = x_train
        Y = y_train
        # Split the data into features and outcome variable=

        # Define the number of features to consider in each subset
        k_values = range(1, len(X.columns) + 1)

        # Define the cross-validation scheme
        cv = RepeatedStratifiedKFold(
            n_splits=inner_splits,
            n_repeats=repeats,
            random_state=1)

        # Initialize variables to store the best feature set and its score
        best_score = 0
        best_roc_auc = 0
        best_features = None

        # Iterate over different feature sets and evaluate their performance
        for k in tqdm(k_values):
            for feature_set in combinations(X.columns, k):
                X_sel = X[list(feature_set)]

                # Evaluate the score using cross-validation
                scores = []
                roc_aucs = []
                for train_idx, test_idx in cv.split(X_sel, Y):
                    X_train, X_test = X_sel.iloc[train_idx], X_sel.iloc[test_idx]
                    y_train, y_test = Y.iloc[train_idx], Y.iloc[test_idx]

                    knn = KNeighborsClassifier()
                    knn.fit(X_train, y_train)
                    score = accuracy_score(y_test, knn.predict(X_test))
                    roc_auc = roc_auc_score(
                        y_test, knn.predict_proba(X_test)[:, 1])
                    scores.append(score)
                    roc_aucs.append(roc_auc)

                # Calculate the mean score across cross-validation repeats
                mean_score = sum(scores) / len(scores)
                mean_roc_auc = sum(roc_aucs) / len(roc_aucs)

                # Update the best feature set if necessary
                if mean_score > best_score:
                    best_score = mean_score
                    best_features = feature_set
                    best_roc_auc = mean_roc_auc

        return list(best_features), best_score, best_roc_auc

    elif select_k:
        X = x_train
        Y = y_train
        k_max = int(len(Y) / 2)

        k_values = range(1, len(X.columns) + 1)

        # Define the cross-validation scheme
        cv = RepeatedStratifiedKFold(
            n_splits=inner_splits,
            n_repeats=repeats,
            random_state=1)

        # Initialize variables to store the best feature set and its score
        best_score = 0
        best_roc_auc = 0
        best_features = None

        # Iterate over different feature sets and evaluate their performance
        for k in tqdm(k_values):
            for feature_set in combinations(X.columns, k):
                X_sel = X[list(feature_set)]

                # Evaluate the score using cross-validation
                scores = []
                roc_aucs = []
                for train_idx, test_idx in cv.split(X_sel, Y):
                    X_train, X_test = X_sel.iloc[train_idx], X_sel.iloc[test_idx]
                    y_train, y_test = Y.iloc[train_idx], Y.iloc[test_idx]

                    knn = KNeighborsClassifier()
                    knn.fit(X_train, y_train)
                    score = accuracy_score(y_test, knn.predict(X_test))
                    roc_auc = roc_auc_score(
                        y_test, knn.predict_proba(X_test)[:, 1])
                    scores.append(score)
                    roc_aucs.append(roc_auc)

                # Calculate the mean score across cross-validation repeats
                mean_score = sum(scores) / len(scores)
                mean_roc_auc = sum(roc_aucs) / len(roc_aucs)

                # Update the best feature set if necessary
                if mean_score > best_score:
                    best_score = mean_score
                    best_features = feature_set
                    best_roc_auc = mean_roc_auc

        neighbour_values = range(1, k_max)

        cv1 = RepeatedStratifiedKFold(
            n_splits=inner_splits,
            n_repeats=5 * repeats,
            random_state=1)

        best_score1 = 0
        best_roc_suc1 = 0
        best_neigh = None
        X_sel = X[list(best_features)]
        for neigh in tqdm(neighbour_values):
            scores1 = []
            roc_aucs1 = []
            for train_idx, test_idx in cv1.split(X, Y):
                X_train, X_test = X_sel.iloc[train_idx], X_sel.iloc[test_idx]
                y_train, y_test = Y.iloc[train_idx], Y.iloc[test_idx]
                knn = KNeighborsClassifier(n_neighbors=neigh)
                knn.fit(X_train, y_train)
                score = accuracy_score(y_test, knn.predict(X_test))
                roc_auc = roc_auc_score(
                    y_test, knn.predict_proba(X_test)[:, 1])
                scores1.append(score)
                roc_aucs1.append(roc_auc)

            mean_score1 = sum(scores1) / len(scores1)
            mean_roc_auc1 = sum(roc_aucs1) / len(roc_aucs1)

            if mean_score1 > best_score1:
                best_score1 = mean_score1
                best_neigh = neigh
                best_roc_auc1 = mean_roc_auc1

        # best_neigh = max(best_neigh, 5)

        return list(best_features), best_score, best_roc_auc, best_neigh


def nested_cv(
        data_file,
        features,
        y_name,
        outname,
        repeats=1000,
        inner_repeats=10,
        outer_splits=3,
        inner_splits=2):
    '''
    This is a function to run nested cross validation. The outer loop is for evaluation and inner loop for feature selection

    Parameters
    ----------

    data_file : dataframe
        pandas dataframe containing input data

    features : list
        total set of features from which features are being selected

    y_name : str
        column name for the outcome variable

    outname : str
         name to be included in the output csv file's name

    outer_splits : int
        number of cross validation splits for the outer loop

    inner_splits : int
        number of cross validation splits for the inner loop

    repeats : int
        number of repeats for the outer loop

    inner_repeats : int
        number of repeats for the inner loop(feature selection)

    Returns
    -------

    file : CSV
        containing cross validation accuracy and ROC AUC for each of the outer loops

    References
    ----------
    - `Scikit-Learn Nested Cross-Validation Example <https://scikit-learn.org/stable/auto_examples/model_selection/plot_nested_cross_validation_iris.html>`__

    - Cawley, Gavin C., and Nicola LC Talbot. "On over-fitting in model selection and subsequent selection bias in performance evaluation." The Journal of Machine Learning Research 11 (2010): 2079-2107.

    '''
    from tqdm import tqdm

    # Load the data into a pandas DataFrame
    df = data_file

    # Split the data into features and outcome variable
    X = df[features]
    y = df[y_name]

    # Define the cross-validation scheme for the outer loop
    outer_cv = RepeatedStratifiedKFold(
        n_splits=outer_splits,
        n_repeats=repeats,
        random_state=1)

    # Initialize variables to store the best feature set and its score
    best_score = 0
    best_features = None
    FEATURES = []
    ACCURACY = []
    ROC_AUC = []
    val_ACC = []
    val_ROC_AUC = []
    # Iterate over the outer cross-validation splits
    for train_idx, test_idx in tqdm(outer_cv.split(X, y)):
        X_train, X_test = X.iloc[train_idx], X.iloc[test_idx]
        y_train, y_test = y.iloc[train_idx], y.iloc[test_idx]

        # Define the cross-validation scheme for the inner loop
        best_features, best_score, best_roc_auc, best_neigh = feature_selection(
            X_train, y_train, features, y_name, inner_splits, select_k=True, repeats=inner_repeats)
        val_ACC.append(best_score)
        val_ROC_AUC.append(best_roc_auc)
        FEATURES.append(best_features)
        print('selected number of neighbours is:', best_neigh)
        knn = KNeighborsClassifier(n_neighbors=best_neigh)
        knn.fit(X_train[best_features], y_train)
        score = accuracy_score(y_test, knn.predict(X_test[best_features]))
        ACCURACY.append(score)
        print('accuracy:', score)
        roc_auc = roc_auc_score(
            y_test, knn.predict_proba(
                X_test[best_features])[
                :, 1])
        ROC_AUC.append(roc_auc)

    DF_out = pd.DataFrame.from_dict({'features': FEATURES,
                                     'validation_ROC_AUC': val_ROC_AUC,
                                     'validation_accuracy': val_ACC,
                                     'test_accuracy': ACCURACY,
                                     'test_ROC_AUC': ROC_AUC})
    DF_out.to_csv('Nested_CV_' + outname + '_fin_results.csv')



import unittest
import numpy as np
from SMdRQA.utils import *

# The assert_matrix function

# Test case class for assert_matrix


class TestAssertMatrix(unittest.TestCase):

    def test_1d_array(self):
        # Test with a 1D array, which should be converted to a 2D array
        input_array = np.array([1, 2, 3])
        result = assert_matrix(input_array)
        self.assertEqual(result.shape, (1, 3))
        np.testing.assert_array_equal(result, np.array([[1, 2, 3]]))

    def test_2d_array(self):
        # Test with a 2D array, which should be returned as is
        input_array = np.array([[1, 2, 3], [4, 5, 6]])
        result = assert_matrix(input_array)
        self.assertEqual(result.shape, (2, 3))
        np.testing.assert_array_equal(result, input_array)

    def test_3d_array(self):
        # Test with a 3D array, which should be returned as is
        input_array = np.array([[[1, 2, 3], [4, 5, 6]]])
        result = assert_matrix(input_array)
        self.assertEqual(result.shape, (1, 2, 3))
        np.testing.assert_array_equal(result, input_array)

    def test_non_numpy_input(self):
        # Test with a non-NumPy array input, which should raise a ValueError
        input_list = [1, 2, 3]
        with self.assertRaises(ValueError):
            assert_matrix(input_list)


class TestCompute3DMatrixSize(unittest.TestCase):

    def test_random_float64_matrix(self):
        # Define matrix dimensions
        dim1, dim2, dim3 = 100, 200, 300

        # Create a random matrix of the given dimensions and dtype float64
        matrix = np.random.rand(dim1, dim2, dim3).astype(np.float64)

        # Compute the expected size in GiB using numpy's nbytes attribute
        expected_size = matrix.nbytes / (1024**3)

        # Compute the size using the function
        result = compute_3D_matrix_size(dim1, dim2, dim3, dtype=np.float64)

        # Assert the two sizes are almost equal (due to floating-point
        # precision)
        self.assertAlmostEqual(result, expected_size, places=6)

    def test_random_float32_matrix(self):
        # Define matrix dimensions
        dim1, dim2, dim3 = 50, 60, 70

        # Create a random matrix of the given dimensions and dtype float32
        matrix = np.random.rand(dim1, dim2, dim3).astype(np.float32)

        # Compute the expected size in GiB using numpy's nbytes attribute
        expected_size = matrix.nbytes / (1024**3)

        # Compute the size using the function
        result = compute_3D_matrix_size(dim1, dim2, dim3, dtype=np.float32)

        # Assert the two sizes are almost equal (due to floating-point
        # precision)
        self.assertAlmostEqual(result, expected_size, places=6)


class TestAssert3DMatrixSize(unittest.TestCase):

    def test_memory_within_limit(self):
        # Test case where the memory requirement should be within the specified
        # limit
        dim1, dim2, dim3 = 100, 100, 100  # Adjust dimensions to be within the 4 GiB limit
        memory_limit = 4  # GiB
        self.assertTrue(
            assert_3D_matrix_size(
                dim1,
                dim2,
                dim3,
                dtype=np.float64,
                memory_limit=memory_limit))

    def test_memory_exceeds_limit(self):
        # Test case where the memory requirement should exceed the specified
        # limit
        dim1, dim2, dim3 = 1000, 1000, 1000  # Dimensions likely to exceed 4 GiB
        memory_limit = 4  # GiB
        self.assertFalse(
            assert_3D_matrix_size(
                dim1,
                dim2,
                dim3,
                dtype=np.float64,
                memory_limit=memory_limit))

    def test_different_dtype_within_limit(self):
        # Test case with a different data type, e.g., float32, which uses less
        # memory
        dim1, dim2, dim3 = 500, 500, 500
        memory_limit = 2  # GiB
        self.assertTrue(
            assert_3D_matrix_size(
                dim1,
                dim2,
                dim3,
                dtype=np.float32,
                memory_limit=memory_limit))

    def test_edge_case_minimal_matrix(self):
        # Edge case with a very small matrix
        dim1, dim2, dim3 = 1, 1, 1
        memory_limit = 0.000001  # GiB
        self.assertTrue(
            assert_3D_matrix_size(
                dim1,
                dim2,
                dim3,
                dtype=np.float64,
                memory_limit=memory_limit))

    def test_large_memory_limit(self):
        # Test case with a large memory limit where any reasonable matrix size
        # would pass
        dim1, dim2, dim3 = 100, 100, 100
        memory_limit = 1000  # GiB
        self.assertTrue(
            assert_3D_matrix_size(
                dim1,
                dim2,
                dim3,
                dtype=np.float64,
                memory_limit=memory_limit))



from SMdRQA.utils import assert_3D_matrix_size
from SMdRQA.utils import compute_3D_matrix_size
from SMdRQA.utils import assert_matrix
from scipy.spatial import distance_matrix
from scipy.special import digamma
from scipy.spatial import distance
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import RepeatedKFold
import memory_profiler
from scipy.interpolate import pchip_interpolate
from functools import partial
from p_tqdm import p_map
from scipy.stats import skew
import random
import pickle
import os
from tqdm import tqdm
import csv
from collections import defaultdict
from os.path import isfile, join
from os import listdir
import pandas as pd
import numpy as np
from scipy.integrate import solve_ivp
import operator
import contextlib
import functools
import operator
import warnings
from numpy.core import overrides
import matplotlib.pyplot as plt
from sklearn.model_selection import RepeatedKFold
import matplotlib
matplotlib.use('Agg')


def find_first_minima_or_global_minima_index(arr):
    n = len(arr)

    if n == 0:
        return None  # Handle empty array

    # Check for the first local minima
    if n == 1 or arr[0] < arr[1]:
        return 0  # Single element or first element is a local minima

    for i in range(1, n - 1):
        if arr[i] < arr[i - 1] and arr[i] < arr[i + 1]:
            return i

    # If no local minima found, return the index of the global minimum
    return np.argmin(arr)


def doanes_formula(data, n):
    '''
    Functions for computing number of bins for data using Doane's formula
    data : input array
    n : number of elements in data
    '''

    g1 = skew(data)
    sigma_g1 = np.sqrt((6 * (n - 2)) / ((n + 1) * (n + 3)))

    k = 1 + np.log2(n) + np.log2(1 + np.abs(g1) / sigma_g1)

    for i in range(len(k)):
        k[i] = int(np.ceil((2 / 3) * k[i]))  # Round up to the nearest integer

    return k


def binscalc(X, n, d, method):
    '''
    Function to calculate number of bins in a multidimensional histogram using a generalised Freedman–Diaconis rule.

    Parameters
    ----------
    x   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    n   : int
        number of samples or observations in the time series

    d   : int
        number of measurements or dimensions of the data

    Returns
    -------

    Bins : array
         number of bins along each dimension

    References
    ----------
    - Freedman, David, and Persi Diaconis. "On the histogram as a density estimator: L 2 theory." Zeitschrift für Wahrscheinlichkeitstheorie und verwandte Gebiete 57.4 (1981): 453-476.
    - Birgé, Lucien, and Yves Rozenholc. "How many bins should be put in a regular histogram." ESAIM: Probability and Statistics 10 (2006): 24-45.

    '''
    if method == 'FD':  # generalised Freedman–Diaconis rule.
        # Generalised Freedman-Diaconis rule with Bins[i] \propto n^(1/(d+2))
        mult_fact = (n**(1 / (2 + d + 10**(-9))))
        Bins = np.zeros(d)
        inf_arr = np.amin(X, axis=0)
        sup_arr = np.amax(X, axis=0)
        q25 = np.quantile(X, 0.25, axis=0)
        q75 = np.quantile(X, 0.75, axis=0)
        IQR = (q75 - q25) + 10**(-9)
        print('IQR:', IQR)
        Bins_1 = np.ceil(mult_fact * (sup_arr - inf_arr) / (2 * IQR))
        Bins = Bins_1.astype(int)
    elif method == 'sqrt':
        ns = [n] * d
        Bins_1 = np.ceil(np.sqrt(ns))
        Bins = Bins_1.astype(int)

    elif method == 'rice':
        ns = [n] * d
        Bins_1 = 2 * np.ceil(np.cbrt(ns))
        Bins = Bins_1.astype(int)
    elif method == 'sturges':
        ns = [n] * d
        Bins_1 = 1 + np.ceil(np.log(ns))
        Bins = Bins_1.astype(int)
    elif method == 'doanes':
        Bins_1 = doanes_formula(X, n)
        Bins = Bins_1.astype(int)

    elif method == 'default':
        Bins = 15

    # print('inf:',inf)
    # print('sup:',sup)
    # print('IQR:',iqr)

    return Bins


def mutualinfo_histdd(X, Y, n, d):
    '''
    Function to calculate mutual information between two time series using multidimensional histogram

    Parameters
    ----------
    x   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    y   : ndarray
        double array of shape (n,d).  second time series

    n   : int
        number of samples or observations in the time series

    d   : int
        number of measurements or dimensions of the data

    Returns
    -------

    MI : double
         mutual information between time series

    References
    ----------
    - Shannon, Claude Elwood. "A mathematical theory of communication." The Bell system technical journal 27.3 (1948): 379-423.

    '''

    points = np.concatenate((X, Y), axis=1)
    bins = binscalc(points, n, 2 * d, 'FD')
    print('BINS:', bins)
    # print('bins:',bins.shape)
    # print('points:',points.shape)
    # 10^-9 added so that x log x does not diverge when x=0 in the calculation
    # of mutual information
    p_xy = np.histogramdd(points, bins=binscalc(
        points, n, 2 * d, 'default'))[0] + 10**(-9)
    p_x = np.histogramdd(X, bins=binscalc(X, n, d, 'default'))[0] + 10**(-9)
    p_y = np.histogramdd(Y, bins=binscalc(Y, n, d, 'default'))[0] + 10**(-9)
    p_xy /= np.sum(p_xy)  # Normalising the probability distribution
    p_x /= np.sum(p_x)
    p_y /= np.sum(p_y)
    return np.sum(p_xy * np.log2(p_xy)) - np.sum(p_x * np.log2(p_x)) - \
        np.sum(p_y * np.log2(p_y))  # formula for mutual information


def mutualinfo_avg(Xmd, Ymd, n, d):
    '''
    Function to calculate mutual information between two time series by avergaring the mutual information across each dimensions

    Parameters
    ----------
    Xmd   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    Ymd   : ndarray
        double array of shape (n,d).  second time series

    n   : int
        number of samples or observations in the time series

    d   : int
        number of measurements or dimensions of the data

    Returns
    -------

    MI : double
         mutual information between time series

    References
    ----------
    - Shannon, Claude Elwood. "A mathematical theory of communication." The Bell system technical journal 27.3 (1948): 379-423.

    '''
    mi = 0
    for i in range(d):
        # Now we can use the same method to compute the mutual information
        X = Xmd[:, i].reshape(-1, 1)
        Y = Ymd[:, i].reshape(-1, 1)
        mi = mi + mutualinfo_histdd(X, Y, n, 1)

    return mi / d


def mutualinfo(X, Y, n, d, method="histdd"):
    '''
    Function to calculate mutual information between two time series

    Parameters
    ----------
    x   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    y   : ndarray
        double array of shape (n,d).  second time series

    n   : int
        number of samples or observations in the time series

    d   : int
        number of measurements or dimensions of the data

    method : Option between computing the mutual information using:
           - multidimensional histogram("histdd")
           - average mutual information across dimensions("avg")

    Returns
    -------

    MI : double
         mutual information between time series

    References
    ----------
    - Shannon, Claude Elwood. "A mathematical theory of communication." The Bell system technical journal 27.3 (1948): 379-423.

    '''
    if method == "histdd":
        mi = mutualinfo_histdd(X, Y, n, d)
    elif method == "avg":
        mi = mutualinfo_avg(X, Y, n, d)

    return mi


# Only aplicable for multidimensional array
def KNN_MI_vectorized(X, Y, nearest_neighbor=5, dtype=np.float64):
    '''
    Function to calculate mutual information between two time series using KNN method for datasets that can't be handled with default binning method. Vectorized version

    Parameters
    ----------
    X   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    Y   : ndarray
        double array of shape (n,d).  second time series

    nearest_neighbor   : int
        number of nearest neighbour for calculating mutual information, default = 5


    Returns
    -------

    MI : double
         mutual information between time series

    References
    ----------
    - Shannon, Claude Elwood. "A mathematical theory of communication." The Bell system technical journal 27.3 (1948): 379-423.
    - Kraskov, A., Stögbauer, H., & Grassberger, P. (2004). Estimating mutual information. Physical Review E—Statistical, Nonlinear, and Soft Matter Physics, 69(6), 066138.

    '''
    X = assert_matrix(X)
    Y = assert_matrix(Y)
    X = X.astype(dtype)  # change the data type to one specified by the user
    Y = Y.astype(dtype)
    n_samples = X.shape[0]

    DX = np.sqrt(
        np.sum(np.square((X[:, np.newaxis, :] - X[np.newaxis, :, :])), axis=2))
    DY = np.sqrt(
        np.sum(np.square((Y[:, np.newaxis, :] - Y[np.newaxis, :, :])), axis=2))
    D_stacked = np.stack((DX, DY), axis=-1)
    D = np.max(D_stacked, axis=2)
    D_sorted = np.sort(D, axis=1)
    # First column would be zero
    k_nearest = np.atleast_2d(D_sorted[:, nearest_neighbor])
    # print('k nearest vectorized:', k_nearest)
    neigh_matrix_X = 1 * ((k_nearest - DX) > 0)
    neigh_X = np.sum(neigh_matrix_X, axis=1) - 1  # Removing "self neighbour"

    neigh_matrix_Y = 1 * ((k_nearest - DY) > 0)
    neigh_Y = np.sum(neigh_matrix_Y, axis=1) - 1  # Removing "self neighbour"
    return digamma(n_samples) + digamma(nearest_neighbor) - \
        np.mean(digamma(neigh_X + 1)) - np.mean(digamma(neigh_Y + 1))


def KNN_MI_partial_vectorized(X, Y, nearest_neighbor=5, dtype=np.float64):
    '''
    Function to calculate mutual information between two time series using KNN method for datasets that can't be handled with default binning method. Partially version. Vectorized version is faster, however, if size of the time series is large and number of dimensions are much larger, the resulting matrix can't be stored in the physical memory of the system (RAM) depending on the resource available. In that case this version can be used to use a relatively faster version compared to non-vectorized version

    Parameters
    ----------
    X   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    Y   : ndarray
        double array of shape (n,d).  second time series

    nearest_neighbor   : int
        number of nearest neighbour for calculating mutual information, default = 5


    Returns
    -------

    MI : double
         mutual information between time series

    References
    ----------
    - Shannon, Claude Elwood. "A mathematical theory of communication." The Bell system technical journal 27.3 (1948): 379-423.
    - Kraskov, A., Stögbauer, H., & Grassberger, P. (2004). Estimating mutual information. Physical Review E—Statistical, Nonlinear, and Soft Matter Physics, 69(6), 066138.

    '''
    X = assert_matrix(X)
    Y = assert_matrix(Y)
    X = X.astype(dtype)  # change the data type to one specified by the user
    Y = Y.astype(dtype)
    XY = np.concatenate((X, Y), axis=1)
    NX = np.zeros(X.shape[0], dtype=int)
    NY = np.zeros(Y.shape[0], dtype=int)
    NXY = np.zeros(XY.shape[0], dtype=int)
    n_samples = X.shape[0]

    for i in range(n_samples):
        # Compute pairwise Euclidean distances
        dist_X = distance.cdist(X[i].reshape(1, -1), X).flatten()
        dist_Y = distance.cdist(Y[i].reshape(1, -1), Y).flatten()

        # Exclude the i-th element (where i == j)
        mask = np.arange(X.shape[0]) != i
        dist_X = dist_X[mask]
        dist_Y = dist_Y[mask]

        # Compute the maximum of distances for each pair (i, j)
        N = sorted(np.maximum(dist_X, dist_Y))

        # Sort the resulting vector
        k_nearest = N[nearest_neighbor - 1]
        NX[i] = np.sum(1 * (dist_X < k_nearest))
        NY[i] = np.sum(1 * (dist_Y < k_nearest))
        # print('k nearest non vectorized:', k_nearest)

    return digamma(n_samples) + digamma(nearest_neighbor) - \
        np.mean(digamma(NX + 1)) - np.mean(digamma(NY + 1))


def KNN_MI_non_vectorized(X, Y, nearest_neighbor=5):
    '''
    Function to calculate mutual information between two time series using KNN method for datasets that can't be handled with default binning method. Non-vectorized version. Vectorized version is faster, however, if size of the time series is large and number of dimensions are much larger, the resulting matrix can't be stored in the physical memory of the system (RAM) depending on the resource available. In that case this version can be used

    Parameters
    ----------
    X   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    Y   : ndarray
        double array of shape (n,d).  second time series

    nearest_neighbor   : int
        number of nearest neighbour for calculating mutual information, default = 5


    Returns
    -------

    MI : double
         mutual information between time series

    References
    ----------
    - Shannon, Claude Elwood. "A mathematical theory of communication." The Bell system technical journal 27.3 (1948): 379-423.
    - Kraskov, A., Stögbauer, H., & Grassberger, P. (2004). Estimating mutual information. Physical Review E—Statistical, Nonlinear, and Soft Matter Physics, 69(6), 066138.

    '''
    XY = np.concatenate((X, Y), axis=1)
    NX = np.zeros(X.shape[0], dtype=int)
    NY = np.zeros(Y.shape[0], dtype=int)
    NXY = np.zeros(XY.shape[0], dtype=int)
    n_samples = X.shape[0]

    for i in range(n_samples):
        N = []
        for j in range(n_samples):
            if i != j:
                N0 = max(
                    distance.euclidean(
                        X[i], X[j]), distance.euclidean(
                        Y[i], Y[j]))
                N.append(N0)
        N.sort()
        k_nearest = N[nearest_neighbor - 1]
        # print('k nearest non vectorized:', k_nearest)

        for j in range(n_samples):
            if i != j:
                if distance.euclidean(X[i], X[j]) < k_nearest:
                    NX[i] += 1
                if distance.euclidean(Y[i], Y[j]) < k_nearest:
                    NY[i] += 1
    print('neigh_X non vectorised:', NX)
    print('neigh_Y non vectorised:', NY)
    return digamma(n_samples) + digamma(nearest_neighbor) - \
        np.mean(digamma(NX + 1)) - np.mean(digamma(NY + 1))


def KNN_MI(
        X,
        Y,
        nearest_neighbor=5,
        method="auto",
        dtype=np.float64,
        memory_limit=4):
    '''
    Function to calculate mutual information between two time series using KNN method for datasets that can't be handled with default binning method. Uses vectorised or non-vectorized version depending on whether the required matrix size is less than the specified memory limit

    Parameters
    ----------
    X   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    Y   : ndarray
        double array of shape (n,d).  second time series

    nearest_neighbor   : int
        number of nearest neighbour for calculating mutual information, default = 5

    method : str
        Specifying options for computing the mutual information using the KNN method. Options are:

        - "auto" : This will check for two additional variables, namely "dtype" (data type) and "memory_limit"
          (maximum memory allocated for the operation). If, for a given matrix size and data type, the vectorized
          algorithm fits within the memory limit, it will proceed with the vectorized version. Otherwise, it will
          compute sequentially.

        - "vectorized" : This will use the vectorized method by default, without checking the memory requirement
          and the limit specified. This option is faster by default.

        - "sequential" : The algorithm is implemented with for loops instead of vectorization. This could be
          significantly slower than the vectorized version. However, if resources (RAM/physical memory) are limited
          and can't handle huge matrices, this option should be chosen.

        - "partial" : best of both worlds between "vectorized" and "sequential"

    dtype   : dtype
        data type, default = np.float64

    memory_limit   : double
        memory limit in GiB, default = 4


    Returns
    -------

    MI : double
         mutual information between time series

    References
    ----------
    - Shannon, Claude Elwood. "A mathematical theory of communication." The Bell system technical journal 27.3 (1948): 379-423.
    - Kraskov, A., Stögbauer, H., & Grassberger, P. (2004). Estimating mutual information. Physical Review E—Statistical, Nonlinear, and Soft Matter Physics, 69(6), 066138.

    '''
    dim1, dim2 = X.shape
    dim3, _ = Y.shape
    pv1 = assert_3D_matrix_size(
        dim1,
        dim2,
        dim3,
        dtype=dtype,
        memory_limit=memory_limit)

    pv2 = assert_3D_matrix_size(
        dim1,
        dim2,
        1,
        dtype=dtype,
        memory_limit=memory_limit)

    if method == "auto":
        if pv1:
            mi = KNN_MI_vectorized(X, Y, nearest_neighbor)
        elif (pv1 == False) and (pv2):
            mi = KNN_MI_partial_vectorized(X, Y, nearest_neighbor, dtype=dtype)
        elif (pv1 == False) and (pv2 == False):
            mi = KNN_MI_non_vectorized(X, Y, nearest_neighbor)

    elif method == "vectorized":
        mi = KNN_MI_vectorized(X, Y, nearest_neighbor, dtype=dtype)

    elif method == "sequential":
        mi = KNN_MI_non_vectorized(X, Y, nearest_neighbor)

    elif method == "partial":
        mi = KNN_MI_partial_vectorized(X, Y, nearest_neighbor, dtype=dtype)

    return mi


def timedelayMI(u, n, d, tau, method="histdd"):
    '''
    Function to calculate mutual information between a time series and a delayed version of itself

    Parameters
    ----------
    u   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    n   : int
        number of samples or observations in the time series

    d   : int
        number of measurements or dimensions of the data

    tau : int
        amount of delay

    Returns
    -------

    MI : double
         mutual information between u and u delayed by tau

    References
    ----------
    - Shannon, Claude Elwood. "A mathematical theory of communication." The Bell system technical journal 27.3 (1948): 379-423.

    '''

    X = u[0:n - tau, :]
    Y = u[tau:n, :]
    return mutualinfo(X, Y, n - tau, d, method=method)


def KNN_timedelayMI(
    u,
    tau,
    nearest_neighbor=5,
    method="auto",
    dtype=np.float64,
        memory_limit=4):
    '''
    Function to calculate mutual information between a time series and a delayed version of itself

    Parameters
    ----------
    u   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    tau : int
        amount of delay

    nearest_neighbor   : int
        number of nearest neighbour for calculating mutual information, default = 5

    method : str
        Specifying options for computing the mutual information using the KNN method. Options are:

        - "auto" : This will check for two additional variables, namely "dtype" (data type) and "memory_limit"
          (maximum memory allocated for the operation). If, for a given matrix size and data type, the vectorized
          algorithm fits within the memory limit, it will proceed with the vectorized version. Otherwise, it will
          compute sequentially.

        - "vectorized" : This will use the vectorized method by default, without checking the memory requirement
          and the limit specified. This option is faster by default.

        - "sequential" : The algorithm is implemented with for loops instead of vectorization. This could be
          significantly slower than the vectorized version. However, if resources (RAM/physical memory) are limited
          and can't handle huge matrices, this option should be chosen.

        - "partial" : best of both worlds between "vectorized" and "sequential"

    dtype   : dtype
        data type, default = np.float64

    memory_limit   : double
        memory limit in GiB, default = 4


    Returns
    -------

    MI : double
         mutual information between time series

    References
    ----------
    - Shannon, Claude Elwood. "A mathematical theory of communication." The Bell system technical journal 27.3 (1948): 379-423.
    - Kraskov, A., Stögbauer, H., & Grassberger, P. (2004). Estimating mutual information. Physical Review E—Statistical, Nonlinear, and Soft Matter Physics, 69(6), 066138.

    '''
    n = u.shape[0]
    X = u[0:n - tau, :]
    Y = u[tau:n, :]
    return KNN_MI(X, Y, nearest_neighbor, method=method,
                  dtype=dtype, memory_limit=memory_limit)


def findtau_default(u, n, d, grp, mi_method="histdd"):
    '''
    Function to calculate correct delay for estimating embedding dimension based on the first minima of the tau vs mutual information curve

    Parameters
    ----------
    u   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    n   : int
        number of samples or observations in the time series

    d   : int
        number of measurements or dimensions of the data

    Returns
    -------

    tau : double
         optimal amount of delay for which mutual information reaches its first minima(and global minima, in case first minima doesn't exist)

    '''

    TAU = []
    MIARR = []
    minMI = timedelayMI(u, n, d, 1, method=mi_method)
    for tau in range(2, n):
        nextMI = timedelayMI(u, n, d, tau, method=mi_method)
        TAU.append(tau)
        MIARR.append(nextMI)
        if nextMI > minMI:
            break
        minMI = nextMI

    return tau - 1


def find_poly_degree(x, y):

    MaxDeg = len(x)
    DEG = []
    RMSE = []
    for deg in range(1, MaxDeg + 1):
        cv = RepeatedKFold(n_splits=10, n_repeats=10, random_state=1)
        MSE_sub = []
        for train_idx, test_idx in cv.split(x, y):
            # Use pandas-style indexing if available, else NumPy indexing.
            if hasattr(x, 'iloc'):
                x_train, x_test = x.iloc[train_idx], x.iloc[test_idx]
                y_train, y_test = y.iloc[train_idx], y.iloc[test_idx]
            else:
                x_train, x_test = x[train_idx], x[test_idx]
                y_train, y_test = y[train_idx], y[test_idx]
            coefficients = np.polyfit(x_train, y_train, deg)
            polynomial = np.poly1d(coefficients)
            y_pred = polynomial(x_test)
            mse = mean_squared_error(y_test, y_pred)
            MSE_sub.append(mse)

        rmse = np.sqrt(np.mean(MSE_sub))
        DEG.append(deg)
        RMSE.append(rmse)

    DEG = np.array(DEG)
    RMSE = np.array(RMSE)
    min_index = np.argmin(RMSE)
    return DEG[min_index]


def findtau_polynomial(u, n, d, grp, mi_method="histdd"):
    '''
    Function to calculate correct delay for estimating embedding dimension based on the first minima of the polynomial fit of tau vs mutual information curve

    Parameters
    ----------
    u   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    n   : int
        number of samples or observations in the time series

    d   : int
        number of measurements or dimensions of the data

    grp : str
        group name for saving the TAU vs MI figure

    Returns
    -------

    tau : double
         optimal amount of delay for which mutual information reaches its first minima(and global minima, in case first minima doesn't exist)

    '''

    TAU = []
    MIARR = []
    for tau in range(2, n):
        nextMI = timedelayMI(u, n, d, tau, method=mi_method)
        TAU.append(tau)
        MIARR.append(nextMI)

    TAU = np.array(TAU)
    MIARR = np.array(MIARR)

    degree = find_poly_degree(TAU, MIARR)

    coefficients = np.polyfit(TAU, MIARR, degree)
    polynomial = np.poly1d(coefficients)
    y_pred = polynomial(TAU)

    tau_index = find_first_minima_or_global_minima_index(y_pred)

    plt.figure()
    plt.plot(TAU, MIARR, 'b*')
    plt.plot(TAU, y_pred, 'r-')
    plt.axvline(x=TAU[tau_index])
    plt.savefig('TAU-MI-' + grp + '.png')

    return TAU[tau_index]


def findtau(u, n, d, grp, method='default', mi_method="histdd"):
    '''
    Function to calculate correct delay for estimating embedding dimension based on either the first minima of the tau vs mutual information curve or the polynomial fit of tau vs mutual information curve


    Parameters
    ----------
    u   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    n   : int
        number of samples or observations in the time series

    d   : int
        number of measurements or dimensions of the data

    method : method that should be used to find the first local minima in MI vs tau curve.
             "default" - first minima of the MI vs TAU plot
             "polynomial"- first minima of the polynomial fit to the MI vs TAU plot

    Returns
    -------

    tau : double
         optimal amount of delay for which mutual information reaches its first minima(and global minima, in case first minima doesn't exist)

    '''

    if method == "default":
        tau = findtau_default(u, n, d, grp, mi_method=mi_method)

    elif method == "polynomial":
        tau = findtau_polynomial(u, n, d, grp, mi_method=mi_method)

    return tau

#### Calculation of m ####################################################


def delayseries(u, n, d, tau, m):
    '''
    Function to calculate correct delay for estimating embedding dimension based on the first minima of the tau vs mutual information curve

    Parameters
    ----------
    u   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    n   : int
        number of samples or observations in the time series

    d   : int
        number of measurements or dimensions of the data

    tau : int
         amount of delay

    m    : int
         number of embedding dimensions

    Returns
    -------

    s : ndarray
       3D matrix ov size ((n-(m-1)*tau,m,d)), time delayed embedded signal

    References
    ----------
    - Takens, F. (1981). Dynamical systems and turbulence. Warwick, 1980, 366–381.

    '''
    s = np.zeros((n - (m - 1) * tau, m, d))
    for i in range(n - (m - 1) * tau):
        for j in range(m):
            s[i, j] = u[i + j * tau]
    return s


def nearest(s, n, d, tau, m):
    '''
    Function that would give a nearest neighbour map, the output array(nn) stores indices of nearest neighbours for index values of each observations

    Parameters
    ----------
    s   : ndarray
        3D matrix ov size ((n-(m-1)*tau,m,d)), time delayed embedded signal

    n   : int
        number of samples or observations in the time series

    d   : int
        number of measurements or dimensions of the data

    tau : int
         amount of delay

    m    : int
         number of embedding dimensions

    Returns
    -------

    nn : array
       an array denoting index of nearest neighbour for each observation

    References
    ----------
    - Takens, F. (1981). Dynamical systems and turbulence. Warwick, 1980, 366–381.

    '''
    nn = np.zeros(n - m * tau, dtype=int)
    nn[0] = n - m * tau - 1
    for i in range(n - m * tau):
        for j in range(n - m * tau):
            if (i != j and np.linalg.norm(
                    s[i] - s[j]) < np.linalg.norm(s[i] - s[nn[i]])):
                nn[i] = j
    return nn


def fnnratio(u, n, d, m, tau, r, sig):
    '''
    Function that calculates the ratio of false nearest neighbours

    Parameters
    ----------
    u   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    n   : int
        number of samples or observations in the time series

    d   : int
        number of measurements or dimensions of the data

    tau : int
         amount of delay

    m    : int
         number of embedding dimensions

    r    : double
         ratio parameter

    sig : double
         standard deviation of the data

    Returns
    -------

    FNN : double
       ratio of false nearest neighbours, for a given embedding dimension m, when compared an embedding dimension m+1

    References
    ----------
    - Kennel, M. B., Brown, R., & Abarbanel, H. D. (1992). Determining embedding dimension for phase-space reconstruction using a geometrical construction. Physical review A, 45 (6), 3403.

    '''
    s1 = delayseries(u, n, d, tau, m)     # embedding in m dimensions
    s2 = delayseries(u, n, d, tau, m + 1)   # embedding in m+1 dimensions
    # containg nearest neghbours after embedding in m dimensions
    nn = nearest(s1, n, d, tau, m)
    isneigh = np.zeros(n - m * tau)
    isfalse = np.zeros(n - m * tau)
    for i in range(n - m * tau):
        disto = np.linalg.norm(s1[i] - s1[nn[i]]) + 10**(-9)
        distp = np.linalg.norm(s2[i] - s2[nn[i]])
        if (disto < sig / r):
            isneigh[i] = 1
            if (distp / disto > r):
                isfalse[i] = 1
    return sum(isneigh * isfalse) / (sum(isneigh) + 10**(-9))


def fnnhitszero(u, n, d, m, tau, sig, delta, Rmin, Rmax, rdiv):
    '''
    Function that finds the value of r at which the FNN ratio can be effectively considered as zero

    Parameters
    ----------
    u   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    n   : int
        number of samples or observations in the time series

    d   : int
        number of measurements or dimensions of the data

    tau : int
         amount of delay

    m    : int
         number of embedding dimensions

    r    : double
         ratio parameter

    sig : double
         standard deviation of the data

    delta: double
         the tolerance value(the maximum difference from zero) for a value of FNN ratio to be effectively considered to be zero

    Rmin : double
         minimum value of r from where we would start the parameter search

    Rmax : double
         maximum value of r for defining the upper limit of parameter search

    rdiv : Int
         number of divisions between Rmin and Rmax for parameter search

    Returns
    -------

    r : double
       value of r at which the value of FNN ratio effectively hits zero

    References
    ----------
    - Kantz, H., & Schreiber, T. (2004). Nonlinear time series analysis (Vol. 7). Cambridge university press. section 3.3.1

    '''
    Rarr = np.linspace(Rmin, Rmax, rdiv)
    for i in range(rdiv):
        if fnnratio(u, n, d, m, tau, Rarr[i], sig) < delta:
            return Rarr[i]
    return -1


def findm(u, n, d, tau, sd, delta, Rmin, Rmax, rdiv, bound):
    '''
    Function that finds the value of m whre the r at which FNN ratio hits zero vs m curve flattens(defined by bound value)

    Parameters
    ----------
    u   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    n   : int
        number of samples or observations in the time series

    d   : int
        number of measurements or dimensions of the data

    tau : int
         amount of delay

    m    : int
         number of embedding dimensions

    r    : double
         ratio parameter

    sig : double
         standard deviation of the data

    delta: double
         the tolerance value(the maximum difference from zero) for a value of FNN ratio to be effectively considered to be zero

    Rmin : double
         minimum value of r from where we would start the parameter search

    Rmax : double
         maximum value of r for defining the upper limit of parameter search

    rdiv : Int
         number of divisions between Rmin and Rmax for parameter search

    bound: double
         bound value for terminating the parameter serch

    Returns
    -------

    m : double
       value of embedding dimension

    References
    ----------
    - Kantz, H., & Schreiber, T. (2004). Nonlinear time series analysis (Vol. 7). Cambridge university press. section 3.3.1

    '''

    mmax = int((3 * d + 11) / 2)
    rm = fnnhitszero(u, n, d, mmax, tau, sd, delta, Rmin, Rmax, rdiv)
    rmp = fnnhitszero(u, n, d, mmax + 1, tau, sd, delta, Rmin, Rmax, rdiv)

    if (rm - rmp > bound):
        return mmax + 1
    for m in range(1, mmax):
        rmp = rm
        rm = fnnhitszero(u, n, d, mmax - m, tau, sd, delta, Rmin, Rmax, rdiv)
        print('rm-rmp:', rm - rmp)
        if (rm - rmp > bound):
            return mmax + 1 - m

    return -1


### Calculation of epsilon ###############################################

def reccplot(u, n, d, m, tau, eps):
    '''
    Function that computes the recurrence plot

    Parameters
    ----------
    u   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    n   : int
        number of samples or observations in the time series

    d   : int
        number of measurements or dimensions of the data

    tau : int
         amount of delay

    m    : int
         number of embedding dimensions

    eps  : double
         radius of the neighbourhood for computing recurrence


    Returns
    -------

    rplot : ndarray
       recurrence plot

    References
    ----------
    - Eckmann, J.-P., Kamphorst, S. O., Ruelle, D., et al. (1995). Recurrence plots of dynamical systems. World Scientific Series on Nonlinear Science Series A, 16, 441–446.

    '''
    # normarr=[]
    s = delayseries(u, n, d, tau, m)
    rplot = np.zeros((n - (m - 1) * tau, n - (m - 1) * tau), dtype=int)
    for i in range(n - (m - 1) * tau):
        for j in range(n - (m - 1) * tau):
            # normarr.append(np.linalg.norm(s[i]-s[j]))
            if np.linalg.norm(s[i] - s[j]) < eps:
                rplot[i, j] = 1
    return rplot


def embedded_signal(
        data=None,
        rdiv=451,
        Rmin=1,
        Rmax=10,
        delta=0.001,
        bound=0.2,
        reqrr=0.1,
        rr_delta=0.005,
        epsmin=0,
        epsmax=10,
        epsdiv=1001):
    (M, N) = data.shape
    data = (data - np.mean(data, axis=0, keepdims=True)) / \
        np.std(data, axis=0, keepdims=True)
    n = M
    d = N
    u = data
    sd = 3 * np.std(u)
    tau = findtau(u, n, d, 0)
    m = findm(u, n, d, tau, sd, delta, Rmin, Rmax, rdiv, bound)
    embedded = delayseries(u, n, d, tau, m)
    return embedded, tau, m


def reccrate(rplot, n):
    '''
    Function that computes the recurrence plot

    Parameters
    ----------
    rplot: ndarray
        recurrence plot

    n   : int
        length of RP


    Returns
    -------

    reccrate : double
       recurrence rate

    References
    ----------
    - Eckmann, J.-P., Kamphorst, S. O., Ruelle, D., et al. (1995). Recurrence plots of dynamical systems. World Scientific Series on Nonlinear Science Series A, 16, 441–446.

    '''
    return float(np.sum(rplot)) / (n * n)


def findeps(u, n, d, m, tau, reqrr, rr_delta, epsmin, epsmax, epsdiv):
    '''
    Function that computes the recurrence plot

    Parameters
    ----------
    u   : ndarray
        multidimensional time series data

    n   : int
        number of observations

    d   : int
        number of dimensions

    tau : int
        amount of delay

    m   : int
        embedding dimension

    reqrr : doubld
        required recurrence rate specified in the input

    rr_delta: double
        tolerance value for considering a value of recurrence rate to be same as the one that is specified in reqrr

    epsmin : double
        lower bound for the parameter search for epsilon(neighbourhood radius)

    epsmax : double
        upper bound for the parameter search for epsilon(neighbourhood radius)

    epsdiv : double
        number of divisions for the parameter search for epsilon(neighbourhood radius) between epsmin and epsmax


    Returns
    -------

    eps   : double
       epsilon(neighbourhood radius)

    References
    ----------
    - Eckmann, J.-P., Kamphorst, S. O., Ruelle, D., et al. (1995). Recurrence plots of dynamical systems. World Scientific Series on Nonlinear Science Series A, 16, 441–446.

    '''
    eps = np.linspace(epsmin, epsmax, epsdiv)
    s = delayseries(u, n, d, tau, m)
    for k in range(epsdiv):
        rplot = np.zeros((n - (m - 1) * tau, n - (m - 1) * tau), dtype=int)
        for i in range(n - (m - 1) * tau):
            for j in range(n - (m - 1) * tau):
                if np.linalg.norm(s[i] - s[j]) < eps[k]:
                    rplot[i, j] = 1
        rr = reccrate(rplot, n - (m - 1) * tau)
        if np.abs(rr - reqrr) < rr_delta:
            return eps[k]

    return -1


def findeps_multi(U, N, D, M, Tau, reqrr, rr_delta, epsmin, epsmax, epsdiv):
    '''
    Function that computes the recurrence plot

    Parameters
    ----------
    U   : list of ndarray
        multidimensional time series data from multiple time series

    N   : list of int
        list of number of observations

    D   : list of int
        list of number of dimensions

    Tau : list of int
        list of amount of delay

    M   : list of int
        list of embedding dimension

    reqrr : doubld
        required recurrence rate specified in the input

    rr_delta: double
        tolerance value for considering a value of recurrence rate to be same as the one that is specified in reqrr

    epsmin : double
        lower bound for the parameter search for epsilon(neighbourhood radius)

    epsmax : double
        upper bound for the parameter search for epsilon(neighbourhood radius)

    epsdiv : double
        number of divisions for the parameter search for epsilon(neighbourhood radius) between epsmin and epsmax


    Returns
    -------

    eps   : double
       epsilon(neighbourhood radius)

    References
    ----------
    - Eckmann, J.-P., Kamphorst, S. O., Ruelle, D., et al. (1995). Recurrence plots of dynamical systems. World Scientific Series on Nonlinear Science Series A, 16, 441–446.

    '''
    num_series = len(N)
    eps = np.linspace(epsmin, epsmax, epsdiv)

    for k in range(epsdiv):
        RR = []
        for item in range(num_series):
            u = U[item]
            n = N[item]
            d = D[item]
            tau = Tau[item]
            m = M[item]
            s = delayseries(u, n, d, tau, m)
            rplot = np.zeros((n - (m - 1) * tau, n - (m - 1) * tau), dtype=int)
            for i in range(n - (m - 1) * tau):
                for j in range(n - (m - 1) * tau):
                    if np.linalg.norm(s[i] - s[j]) < eps[k]:
                        rplot[i, j] = 1
            rr_sub = reccrate(rplot, n - (m - 1) * tau)
            RR.append(rr_sub)
        rr = np.mean(RR)
        if np.abs(rr - reqrr) < rr_delta:
            return eps[k]

    return -1

### Calculation of RQA parameters ########################################


def plotwindow(M, n, win, i, j):
    window = np.zeros((win, win))
    for a in range(win):
        for b in range(win):
            window[a, b] = M[i + a, j + b]
    return window


def vert_hist(M, n):  # Functio to calculate vertical line distribution
    '''
    Function to compute vertical line distribution(counts of line lengths)

    Parameters
    ----------
    m   : ndarray
        recurrence plot

    n   : int
        length of RP

    Returns
    -------

    nvert : array
       an array containing counts of line lengths

    References
    ----------
    - Webber Jr, C. L., & Zbilut, J. P. (1994). Dynamical assessment of physiological systems and states using recurrence plot strategies. Journal of applied physiology, 76 (2), 965–973.
    - Webber Jr, C. L., & Zbilut, J. P. (2005). Recurrence quantification analysis of nonlinear dynamical systems. Tutorials in contemporary nonlinear methods for the behavioral sciences, 94 (2005), 26–94.
    - Marwan, N., Romano, M. C., Thiel, M., & Kurths, J. (2007). Recurrence plots for the analysis of complex systems. Physics reports, 438 (5-6), 237–329.
    - Marwan, N., Schinkel, S., & Kurths, J. (2013). Recurrence plots 25 years later—gaining confidence in dynamical transitions. Europhysics Letters, 101 (2), 20007.
    - Marwan, N., Wessel, N., Meyerfeldt, U., Schirdewan, A., & Kurths, J. (2002). Recurrence- plot-based measures of complexity and their application to heart-rate-variability data. Physical review E, 66 (2), 026702.

    '''
    nvert = np.zeros(n + 1)
    for i in range(n):
        counter = 0
        for j in range(n):
            if M[j][i] == 1:
                counter += 1
            else:
                nvert[int(counter)] += 1
                counter = 0
        nvert[counter] += 1
    return nvert


def onedhist(M, n):
    hst = np.zeros(n + 1)
    counter = 0
    for i in range(n):
        if M[i] == 1:
            counter += 1
        else:
            hst[counter] += 1
            counter = 0
    hst[counter] += 1
    return hst


def diaghist(M, n):  # Function to calculate diagonal line distribution
    '''
    Function to compute diagonal line distribution(counts of line lengths)

    Parameters
    ----------

    m   : ndarray
        recurrence plot

    n   : int
        length of RP

    Returns
    -------

    nvert : array
       an array containing counts of line lengths

    References
    ----------
    - Webber Jr, C. L., & Zbilut, J. P. (1994). Dynamical assessment of physiological systems and states using recurrence plot strategies. Journal of applied physiology, 76 (2), 965–973.
    - Webber Jr, C. L., & Zbilut, J. P. (2005). Recurrence quantification analysis of nonlinear dynamical systems. Tutorials in contemporary nonlinear methods for the behavioral sciences, 94 (2005), 26–94.
    - Marwan, N., Romano, M. C., Thiel, M., & Kurths, J. (2007). Recurrence plots for the analysis of complex systems. Physics reports, 438 (5-6), 237–329.
    - Marwan, N., Schinkel, S., & Kurths, J. (2013). Recurrence plots 25 years later—gaining confidence in dynamical transitions. Europhysics Letters, 101 (2), 20007.
    - Marwan, N., Wessel, N., Meyerfeldt, U., Schirdewan, A., & Kurths, J. (2002). Recurrence- plot-based measures of complexity and their application to heart-rate-variability data. Physical review E, 66 (2), 026702.

    '''
    dghist = np.zeros(n + 1)
    for i in range(n):
        diag = np.zeros(n - i)
        for j in range(n - i):
            diag[j] = M[i + j][j]
        subdiaghist = onedhist(diag, n - i)
        for k in range(n - i + 1):
            dghist[k] += subdiaghist[k]
    dghist *= 2
    dghist[n] /= 2
    return dghist


### Measures to capture probability distributions ########################

def percentmorethan(hst, mini, n):
    '''
    Function to compute determinism and laminarity from the histogram distribution of lines

    The idea is to see what fraction of rcurrent points are part of a linear structure, which can be either
    vertical or horizontal. The definition of a linear structure on an RP is based on the minimum length(mini)
    given as an input

    Parameters
    ----------
    hst   : array
        histogram counts of line lengths

    mini  : int
        minimum length of consecutive occurances of value 1 in the RP(either vertically or horizontally) that is considered as a line

    n   : int
        length of RP

    Returns
    -------

    nvert : array
       an array containing counts of line lengths

    References
    ----------
    - Webber Jr, C. L., & Zbilut, J. P. (1994). Dynamical assessment of physiological systems and states using recurrence plot strategies. Journal of applied physiology, 76 (2), 965–973.
    - Webber Jr, C. L., & Zbilut, J. P. (2005). Recurrence quantification analysis of nonlinear dynamical systems. Tutorials in contemporary nonlinear methods for the behavioral sciences, 94 (2005), 26–94.
    - Marwan, N., Romano, M. C., Thiel, M., & Kurths, J. (2007). Recurrence plots for the analysis of complex systems. Physics reports, 438 (5-6), 237–329.
    - Marwan, N., Schinkel, S., & Kurths, J. (2013). Recurrence plots 25 years later—gaining confidence in dynamical transitions. Europhysics Letters, 101 (2), 20007.
    - Marwan, N., Wessel, N., Meyerfeldt, U., Schirdewan, A., & Kurths, J. (2002). Recurrence- plot-based measures of complexity and their application to heart-rate-variability data. Physical review E, 66 (2), 026702.

    '''
    numer = 0
    denom = 10**(-7)
    for i in range(mini, n + 1):
        numer += i * hst[i]
    for i in range(1, n + 1):
        denom += i * hst[i]
    return numer / denom


def mode(hst, mini, n):
    '''
    Function to find mode of the line distributions

    Parameters
    ----------
    hst   : array
        histogram counts of line lengths

    mini  : int
        minimum length of consecutive occurances of value 1 in the RP(either vertically or horizontally) that is considered as a line

    n   : int
        length of RP

    Returns
    -------

    p : int
       mode of line length distribution

    References
    ----------
    - Webber Jr, C. L., & Zbilut, J. P. (1994). Dynamical assessment of physiological systems and states using recurrence plot strategies. Journal of applied physiology, 76 (2), 965–973.
    - Webber Jr, C. L., & Zbilut, J. P. (2005). Recurrence quantification analysis of nonlinear dynamical systems. Tutorials in contemporary nonlinear methods for the behavioral sciences, 94 (2005), 26–94.
    - Marwan, N., Romano, M. C., Thiel, M., & Kurths, J. (2007). Recurrence plots for the analysis of complex systems. Physics reports, 438 (5-6), 237–329.
    - Marwan, N., Schinkel, S., & Kurths, J. (2013). Recurrence plots 25 years later—gaining confidence in dynamical transitions. Europhysics Letters, 101 (2), 20007.
    - Marwan, N., Wessel, N., Meyerfeldt, U., Schirdewan, A., & Kurths, J. (2002). Recurrence- plot-based measures of complexity and their application to heart-rate-variability data. Physical review E, 66 (2), 026702.

    '''
    p = mini
    for i in range(mini + 1, n + 1):
        if hst[i] > hst[p]:
            p = i
    return p


def maxi(hst, mini, n):
    '''
    maximum value in the line length distribution

    Parameters
    ----------
    hst   : array
        histogram counts of line lengths

    mini  : int
        minimum length of consecutive occurances of value 1 in the RP(either vertically or horizontally) that is considered as a line

    n   : int
        length of RP

    Returns
    -------

    plmax : int
       max of line length distribution

    References
    ----------
    - Webber Jr, C. L., & Zbilut, J. P. (1994). Dynamical assessment of physiological systems and states using recurrence plot strategies. Journal of applied physiology, 76 (2), 965–973.
    - Webber Jr, C. L., & Zbilut, J. P. (2005). Recurrence quantification analysis of nonlinear dynamical systems. Tutorials in contemporary nonlinear methods for the behavioral sciences, 94 (2005), 26–94.
    - Marwan, N., Romano, M. C., Thiel, M., & Kurths, J. (2007). Recurrence plots for the analysis of complex systems. Physics reports, 438 (5-6), 237–329.
    - Marwan, N., Schinkel, S., & Kurths, J. (2013). Recurrence plots 25 years later—gaining confidence in dynamical transitions. Europhysics Letters, 101 (2), 20007.
    - Marwan, N., Wessel, N., Meyerfeldt, U., Schirdewan, A., & Kurths, J. (2002). Recurrence- plot-based measures of complexity and their application to heart-rate-variability data. Physical review E, 66 (2), 026702.

    '''
    lmax = 1
    for i in range(1, n - 1):
        if hst[i] != 0:
            lmax = i
    return lmax


def average(hst, mini, n):
    '''
    Function to find mean of the line distributions

    Parameters
    ----------
    hst   : array
        histogram counts of line lengths

    mini  : int
        minimum length of consecutive occurances of value 1 in the RP(either vertically or horizontally) that is considered as a line

    n   : int
        length of RP

    Returns
    -------

    mu : double
       mean of line length distribution

    References
    ----------
    - Webber Jr, C. L., & Zbilut, J. P. (1994). Dynamical assessment of physiological systems and states using recurrence plot strategies. Journal of applied physiology, 76 (2), 965–973.
    - Webber Jr, C. L., & Zbilut, J. P. (2005). Recurrence quantification analysis of nonlinear dynamical systems. Tutorials in contemporary nonlinear methods for the behavioral sciences, 94 (2005), 26–94.
    - Marwan, N., Romano, M. C., Thiel, M., & Kurths, J. (2007). Recurrence plots for the analysis of complex systems. Physics reports, 438 (5-6), 237–329.
    - Marwan, N., Schinkel, S., & Kurths, J. (2013). Recurrence plots 25 years later—gaining confidence in dynamical transitions. Europhysics Letters, 101 (2), 20007.
    - Marwan, N., Wessel, N., Meyerfeldt, U., Schirdewan, A., & Kurths, J. (2002). Recurrence- plot-based measures of complexity and their application to heart-rate-variability data. Physical review E, 66 (2), 026702.

    '''
    numer = 0
    denom = 10**(-7)
    for i in range(mini, n + 1):
        numer += i * hst[i]
        denom += hst[i]
    return numer / denom


def entropy(hst, mini, n):
    '''
    Function to find entropy of the line distributions

    Parameters
    ----------
    hst   : array
        histogram counts of line lengths

    mini  : int
        minimum length of consecutive occurances of value 1 in the RP(either vertically or horizontally) that is considered as a line

    n   : int
        length of RP

    Returns
    -------

    ent : int
       entropy of line length distribution

    References
    ----------
    - Webber Jr, C. L., & Zbilut, J. P. (1994). Dynamical assessment of physiological systems and states using recurrence plot strategies. Journal of applied physiology, 76 (2), 965–973.
    - Webber Jr, C. L., & Zbilut, J. P. (2005). Recurrence quantification analysis of nonlinear dynamical systems. Tutorials in contemporary nonlinear methods for the behavioral sciences, 94 (2005), 26–94.
    - Marwan, N., Romano, M. C., Thiel, M., & Kurths, J. (2007). Recurrence plots for the analysis of complex systems. Physics reports, 438 (5-6), 237–329.
    - Marwan, N., Schinkel, S., & Kurths, J. (2013). Recurrence plots 25 years later—gaining confidence in dynamical transitions. Europhysics Letters, 101 (2), 20007.
    - Marwan, N., Wessel, N., Meyerfeldt, U., Schirdewan, A., & Kurths, J. (2002). Recurrence- plot-based measures of complexity and their application to heart-rate-variability data. Physical review E, 66 (2), 026702.

    '''
    summ = 0
    entr = 0
    for i in range(mini, n + 1):
        summ += hst[i]
    for i in range(mini, n + 1):
        if (hst[i] != 0):
            entr -= (hst[i] / summ) * np.log(hst[i] / summ)
    return entr



import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import scipy.stats as ss
from nolitsa import d2, dimension, surrogates
import numpy as np
from scipy.fftpack import fft, ifft
from scipy.signal import find_peaks
from scipy.interpolate import interp1d
from numpy.random import rand, randn
from datetime import datetime
from SMdRQA.RQA_functions import embedded_signal
from scipy import signal
import random
import numba


def wrapTo2Pi(x):
    xwrap = np.remainder(x, 2 * np.pi)
    mask = np.abs(xwrap) > np.pi
    xwrap[mask] -= 2 * np.pi * np.sign(xwrap[mask])
    mask1 = x < 0
    mask2 = np.remainder(x, np.pi) == 0
    mask3 = np.remainder(x, 2 * np.pi) != 0
    xwrap[mask1 & mask2 & mask3] -= 2 * np.pi
    return xwrap


def Chi2_test(signal):
    '''
    Stationarity test for time series data

    The stationarity test proposed by Isliker & Kurths (1993)
    compares the distribution of the first half of the time series with
    the distribution of the entire time series using a chi-squared test.
    If there is no significant difference in the two distributions, we
    can consider the time series to be stationary.

    Parameters
    ----------
    signal : array
        One diensional time series signal

    Returns
    -------

    chisq : float
        Chi-square statistic value

    p : float
        p-vaue for chi-square test

    References
    ----------
    - Isliker, Heinz, and Juergen Kurths. "A test for stationarity: finding parts in time series apt for correlation dimension estimates.
      " International Journal of Bifurcation and Chaos 3.06 (1993): 1573-1579.
    - Mannattil, Manu, Himanshu Gupta, and Sagar Chakraborty. "Revisiting evidence of chaos in X-ray light curves: the case of GRS 1915+ 105."
      The Astrophysical Journal 833.2 (2016): 208.

    '''

    N = len(signal)
    reminder = N % 2
    if reminder == 0:
        Nhalf = int(N / 2)
        first_half = signal[0:Nhalf]

    elif reminder == 1:
        Nhalf = int(np.floor(N / 2) + 1)
        first_half = signal[0:Nhalf]

    nq, edges = np.histogram(signal)
    nqhalf, edges2 = np.histogram(first_half, bins=edges)
    Expected = ((Nhalf / N) * nq) + (10 ** (-9))
    (chisq, p) = ss.chisquare(f_obs=nqhalf, f_exp=Expected)

    return chisq, p


def Correlation_sum_plot(
    sig=None, r=100, metric="chebyshev", window=10, save_name=None
):
    '''
    Compute the correlation sum plots.

    Computes the correlation sum of the given time series for the specified distances (Grassberger & Procaccia 1983).

    Parameters
    ----------
    sig : ndarray
        N-dimensional real input array containing points in the phase space.
    r : int or array, optional (default = 100)
        Distances for which the correlation sum should be calculated.
        If r is an int, then the distances are taken to be a geometric progression between a minimum and maximum length scale
        (estimated according to the metric and the input series).
    metric : string, optional (default = 'chebyshev')
        Metric to use for distance computation.  Must be one of
        "chebyshev" (aka the maximum norm metric), "cityblock" (aka the Manhattan metric), or "euclidean".
    window : int, optional (default = 10)
        Minimum temporal separation (Theiler window) that should exist between pairs.
    save_path: str(optional)
        File name for saving the figure

    Returns
    -------
    figure

    References
    ----------
    - Grassberger, Peter, and Itamar Procaccia. "Characterization of strange attractors." Physical review letters 50.5 (1983): 346.
    - Mannattil, Manu, Himanshu Gupta, and Sagar Chakraborty. "Revisiting evidence of chaos in X-ray light curves: the case of GRS 1915+ 105." The Astrophysical Journal 833.2 (2016): 208.
    '''
    r, c = d2.c2(sig, r=r, metric=metric, window=window)
    plt.figure(figsize=(16, 12))
    plt.plot(r, c, "b-")
    plt.title("Correlation Sum Plot")
    plt.xlabel("r")
    plt.ylabel("C(r)")
    if save_name:
        plt.savefig(save_name)
    else:
        plt.show()


def time_irreversibility(sig):
    '''
    Compute the time irriversibility .

    Computes the time irriversibility of the signal

    Parameters
    ----------
    sig : ndarray
        N-dimensional real input array containing points in the phase
        space.

    Returns
    -------
    TI : float
       time irriversibility of the signal

    References
    ----------
    - Discrimination power of measures for nonlinearity in a time series. Physical Review E, 55 (5), 5443.


    '''
    before = sig[:-1]
    after = sig[1:]
    array = (after - before) ** 3
    TI = (1 / len(sig)) * np.sum(array)
    return TI


def sort_matlab(arr, dim=None):
    '''
    Sorts the elements of an array along a specified dimension in ascending order,
    similar to MATLAB's sort function.

    Parameters:
    arr : numpy.ndarray
        The input array to be sorted.

    dim : int, optional
        The dimension along which to sort the array. Default is None, which means
        the array is flattened before sorting.

    Returns:
    sorted_arr : numpy.ndarray
        The sorted array.

    sorted_indices : numpy.ndarray
        The indices that would sort the original array.
    '''
    if dim is not None:
        sorted_indices = np.argsort(arr, axis=dim)
        sorted_arr = np.take_along_axis(arr, sorted_indices, axis=dim)
    else:
        flattened_indices = np.argsort(arr, axis=None)
        sorted_arr = np.take_along_axis(arr, flattened_indices, axis=None)
        sorted_indices = np.unravel_index(flattened_indices, arr.shape)

    return sorted_arr, sorted_indices


def preprocessing(sig, fs):
    '''
    pre processing for surrogate data generation

    function is used to truncate the input signal based on minimizing the mismatch between consecutive points at
    the beginning and end of the signal, along with some initial processing steps like    mean subtraction and
    time vector generation. Adapted from Gemma, et al.,2018

    Parameters
    ----------
    sig : ndarray
        1-dimensional real input array containing points in the phase
        space.

    fs : int
       sampling rate of the signal

    Returns
    -------
    cutsig : ndarray
       1-dimensional array of truncated signal

    t2: float
       total length(in terms of time) of the truncated signal

    kstart : float
       starting point of the truncated signal w.r.t the original signal

    kend : float
       ending point of the truncated signal w.r.t the original signal

    References
    ----------
    - Lancaster, Gemma, et al. "Surrogate data for hypothesis testing of physical systems." Physics Reports 748 (2018): 1-60.


    '''
    sig = sig - np.mean(sig)
    t = np.linspace(0, len(sig) / fs, len(sig))
    L = len(sig)
    p = 10  # Find pair of points which minimizes mismatch between p consecutive
    # points and the beginning and the end of the signal
    K1 = round(L / 100)  # Proportion of signal to consider at the beginning
    k1 = sig[:K1]
    K2 = round(L / 10)  # Proportion of signal to consider at the end
    k2 = sig[-K2:]
    # Truncate to match start and end points and first derivatives
    if len(k1) <= p:
        p = len(k1) - 1
    d = np.zeros((len(k1) - p, len(k2) - p))
    for j in range(len(k1) - p):
        for k in range(len(k2) - p):
            d[j, k] = sum(abs(k1[j: j + p] - k2[k: k + p]))
    v, I = min(abs(d), axis=1)
    I2 = np.argmin(v)  # Minimum mismatch
    kstart = I2
    kend = I[I2] + len(sig[:-K2])
    if kend > len(sig):
        kend = len(sig)
    cutsig = sig[kstart:kend]  # New truncated time series
    t2 = t[kstart:kend]  # Corresponding time
    return cutsig, t2, kstart, kend


def surrogate(sig, N, method, pp, fs, *args):
    '''
    function for generating surrogate data

    function is used to generate surrogate data for hypothesis testing. Adapted from Gemma, et al.,2018

    Parameters
    ----------
    sig : ndarray
        1-dimensional real input array containing points in the phase
        space.

    N  : int
       Number of surrogates to calculate



    fs : int
       sampling rate of the signal

    pp : bool(default = True)
       preprocessing on (True) or off (False)

    method : str
       Method used for generating surrogates

       *RP* : Random permutation surrogates

       *FT* : Fourier transform surrogates

       *AAFT* : Amplitude adjusted Fourier transform

       *IAAFT1* : Iterative amplitude adjusted Fourier transform with exact distribution

       *IAAFT2* : Iterative amplitude adjusted Fourier transform with exact spectrum

       *CPP* : Cyclic phase permutation

       *PPS* : Pseudo-periodic

       *TS* : Twin

       *tshift* : Time shifted

       *CSS* : Cycle shuffled surrogates. Require that the signal can be
         separated into distinct cycles. May require adjustment of peak finding
         parameters.


    Returns
    -------
    surr : ndarray
       surrogate data

    params : dict
       parameters from the generation, including truncation locations if preprocessed and runtime.

    References
    ----------
    - Lancaster, Gemma, et al. "Surrogate data for hypothesis testing of physical systems." Physics Reports 748 (2018): 1-60.
    - Theiler, James, et al. "Testing for nonlinearity in time series: the method of surrogate data." Physica D: Nonlinear Phenomena 58.1-4 (1992): 77-94.
    - Schreiber, Thomas, and Andreas Schmitz. "Improved surrogate data for nonlinearity tests." Physical review letters 77.4 (1996): 635.
    - Small, Michael, Dejin Yu, and Robert G. Harrison. "Surrogate test for pseudoperiodic time series data." Physical Review Letters 87.18 (2001): 188101.
    - Thiel, Marco, et al. "Twin surrogates to test for complex synchronisation." Europhysics Letters 75.4 (2006): 535.
    - Theiler, James. "On the evidence for low-dimensional chaos in an epileptic electroencephalogram." Physics Letters A 196.1-2 (1994): 335-341.


    '''
    origsig = sig
    params = {}
    params["origsig"] = origsig
    params["method"] = method
    params["numsurr"] = N
    params["fs"] = fs
    z = datetime.now()
    # Preprocessing
    if pp:
        sig, time, ks, ke = preprocessing(sig, fs)
    else:
        time = np.linspace(0, len(sig) / fs, len(sig))
    L = len(sig)
    L2 = np.ceil(L / 2)
    if pp == 1:
        params["preprocessing"] = "on"
        params["cutsig"] = sig
        params["sigstart"] = ks
        params["sigend"] = ke
    else:
        params["preprocessing"] = "off"
    params["time"] = time
    # Random permutation (RP) surrogates
    if method == "RP":
        surr = np.zeros((N, len(sig)))
        for k in range(N):
            surr[k, :] = sig[np.random.permutation(L)]
    # Fourier transform (FT) surrogate
    elif method == "FT":
        a = 0
        b = 2 * np.pi
        if len(args) > 0:
            eta = args[0]
        else:
            eta = (b - a) * rand(N, L2 - 1) + a  # Random phases
        ftsig = fft(sig)  # Fourier transform of signal
        ftrp = np.zeros((N, len(ftsig)))
        ftrp[:, 0] = ftsig[0]
        F = ftsig[1:L2]
        F = F[np.ones(N), :]
        ftrp[:, 1:L2] = F * (np.exp(1j * eta))
        ftrp[:, 2 + L - L2: L] = np.conj(np.fliplr(ftrp[:, 2:L2]))
        surr = ifft(ftrp, axis=1)
        params["rphases"] = eta
    # Amplitude adjusted Fourier transform surrogate
    elif method == "AAFT":
        a = 0
        b = 2 * np.pi
        eta = (b - a) * np.random.randn(N, L2 - 1) + a  # Random phases
        val, ind = sort_matlab(sig)
        rankind = np.zeros(L)
        rankind[ind] = range(L)  # Rank the locations
        gn = sort_matlab(
            np.random.randn(N, len(sig)), dim=1
        )  # Create Gaussian noise signal and np.np.sort
        for j in range(N):
            gn[j, :] = gn[
                j, rankind
            ]  # Reorder noise signal to match ranks in original signal
        ftgn = fft(gn, axis=1)
        F = ftgn[:, 1:L2]
        surr = np.zeros((N, len(sig)))
        surr[:, 0] = gn[:, 0]
        surr[:, 1:L2] = F * np.exp(1j * eta)
        surr[:, 2 + L - L2: L] = np.conj(np.flipr(surr[:, 2:L2]))
        surr = ifft(surr, axis=1)
        _, ind2 = sort_matlab(surr, dim=1)  # Sort surrogate
        rrank = np.zeros(L)
        for k in range(N):
            rrank[ind2[k, :]] = range(L)
            surr[k, :] = val[rrank]
    # Iterated amplitude adjusted Fourier transform (IAAFT-1) with exact
    # distribution
    elif method == "IAAFT1":
        maxit = 1000
        val, ind = sort_matlab(sig)  # Sorted list of values
        rankind = np.zeros(L)  # Rank the values
        rankind[ind] = range(L)
        ftsig = fft(sig)
        F = ftsig[np.ones(N), :]
        surr = np.zeros((N, L))
        it = 1
        irank = rankind[np.ones(N), :]
        irank2 = np.zeros(L)
        oldrank = np.zeros((N, L))
        iind = np.zeros((N, L))
        iterf = np.zeros((N, L))
        while max(max(abs(oldrank - irank), axis=1)) != 0 and it < maxit:
            go = max(abs(oldrank - irank), axis=1)
            inc = np.where(go != 0)[0]
            oldrank = irank
            iterf[inc, :] = np.real(
                ifft(
                    abs(F[inc, :]) * np.exp(1j * np.angle(fft(surr[inc, :], axis=1))), axis=1
                )
            )
            _, iind[inc, :] = sort_matlab(iterf[inc, :], dim=1)
            for k in inc:
                irank2[iind[k, :]] = range(L)
                irank[k, :] = irank2
                surr[k, :] = val[irank2]
            it += 1
    # Iterated amplitude adjusted Fourier transform (IAAFT-2) with exact
    # spectrum
    elif method == "IAAFT2":
        maxit = 1000
        val, ind = sort_matlab(sig)  # Sorted list of values
        rankind = np.zeros(L)  # Rank the values
        rankind[ind] = range(L)
        ftsig = fft(sig)
        F = ftsig[np.ones(N), :]
        surr = np.zeros((N, L))
        it = 1
        irank = rankind[np.ones(N), :]
        irank2 = np.zeros(L)
        oldrank = np.zeros((N, L))
        iind = np.zeros((N, L))
        iterf = np.zeros((N, L))
        while max(max(abs(oldrank - irank), axis=1)) != 0 and it < maxit:
            go = max(abs(oldrank - irank), axis=1)
            inc = np.where(go != 0)[0]
            oldrank = irank
            iterf[inc, :] = np.real(
                ifft(
                    abs(F[inc, :]) * np.exp(1j * np.angle(fft(surr[inc, :], axis=1))), axis=1
                )
            )
            _, iind[inc, :] = sort_matlab(iterf[inc, :], dim=1)
            for k in inc:
                irank2[iind[k, :]] = range(L)
                irank[k, :] = irank2
                surr[k, :] = val[irank2]
            it += 1
        surr = iterf
    # Cyclic phase permutation (CPP) surrogates
    elif method == "CPP":
        phi = wrapTo2Pi(sig)
        pdiff = phi[1:] - phi[:-1]
        locs = np.where(pdiff < -np.pi)[0]
        parts = []
        for j in range(len(locs) - 1):
            tsig = phi[locs[j] + 1: locs[j + 1]]
            parts.append(tsig)
        st = phi[: locs[0]]
        en = phi[locs[-1] + 1:]
        surr = np.zeros((N, L))
        for k in range(N):
            surr[k, :] = np.unwrap(
                np.hstack((st, parts[np.random.permutation(len(parts))], en)))
    # Pseudo-periodic surrogates (PPS)
    elif method == "PPS":
        # Embedding of original signal
        sig, tau, m = embedded_signal(data=np.reshape(sig, (len(sig), 1)))
        L = len(sig)
        L2 = np.ceil(L / 2)
        time = np.linspace(0, len(sig) / fs, len(sig))
        params["embed_delay"] = tau
        params["embed_dim"] = m
        params["embed_sig"] = sig
        # Find the index of the first nearest neighbour from the first half of the
        # embedded signal to its last value to avoid cycling near last value
        matr = max(abs(sig[:, :] - sig[:, k] * np.ones(1, L))
                   for k in range(L))
        ssig, mind = min(matr[matr > 0], axis=1)
        _, pl = min(matr[: round(L / 2), L])
        rho = 0.7 * np.mean(ssig)
        for x in range(N):
            kn = random.randint(L)  # Choose random starting point
            for j in range(
                L
            ):  # Length of surrogate is the same as the embedded time series
                if kn == L:
                    kn = pl
                kn += 1  # Move forward from previous kn
                # Set surrogate to current value for kn (choose first
                # component, can be any)
                surr[x, j] = sig[0, kn]
                sigdist = max(
                    abs(sig[:, :] - sig[:, kn] * np.ones(1, L))
                )  # Find the maximum
                # distance between each point in the original signal and the current
                # values with noise added
                _, kn = min(sigdist)  # Find nearest point
    # Twin surrogates
    elif method == "TS":
        # Embedding of original signal
        sig, tau, m = embedded_signal(data=np.reshape(sig, (len(sig), 1)))
        L = len(sig)
        L2 = np.ceil(L / 2)
        time = np.linspace(0, len(sig) / fs, len(sig))
        params["embed_delay"] = tau
        params["embed_dim"] = m
        params["embed_sig"] = sig
        dL = L
        alpha = 0.1
        Rij = np.zeros((L, L))
        for k in range(1, L):
            Rij[k, :k] = max(abs(sig[:, :k] - sig[:, k] * np.ones(1, k)))
        Rij = Rij + Rij.T
        _, pl = min(Rij[: round(L / 2), L])
        Sij = sort_matlab(Rij.flatten())
        delta = Sij[round(alpha * L**2)]
        Rij[Rij < delta] = -1
        Rij[Rij > delta] = 0
        Rij = abs(Rij)
        ind = [[] for _ in range(L)]
        eln = np.zeros(L)
        twind = np.arange(L)
        remp = np.arange(1, L + 1)  # remaining points
        while len(remp) > 0:
            twn = remp[0]
            ind[twn] = remp[max(
                abs(Rij[:, remp] - Rij[:, twn] * np.ones(1, len(remp)))) == 0]
            ind[ind[twn]] = ind[twn]
            eln[ind[twn]] = len(ind[twn])
            twind[ind[twn]] = 0
            remp = twind[twind > 0]
        for sn in range(N):
            kn = random.randint(L) - 1
            for j in range(dL):
                kn += 1
                surr[sn, j] = sig[0, kn]
                kn = ind[kn][random.randint(eln[kn])]
                if kn == L:
                    kn = pl
    # Time-shifted surrogates
    elif method == "tshift":
        for sn in range(N):
            startp = random.randint(L - 1)
            surr[sn, :] = np.hstack((sig[1 + startp: L], sig[1:startp]))
    # Cycle shuffled surrogates
    elif method == "CSS":
        if len(args) > 0:
            MPH = args[0]  # Minimum heak height
            MPD = args[1]  # Minimum peak distance
        else:
            MPH = 0
            MPD = fs
        I = signal.find_peaks(sig, height=MPH, distance=MPD)
        st = sig[: I[0] - 1]
        en = sig[I[-1]:]
        parts = [sig[I[j]: I[j + 1] - 1] for j in range(len(I) - 1)]
        for k in range(N):
            surr[k, :] = np.unwrap(
                np.hstack((st, parts[np.random.permutation(len(parts))], en)))
    params["runtime"] = (datetime.now() - z).total_seconds()
    params["type"] = method
    return surr, params



from scipy.spatial import distance_matrix
from scipy.special import digamma
from scipy.spatial import distance
import memory_profiler
from scipy.interpolate import pchip_interpolate
from functools import partial
from p_tqdm import p_map
from scipy.stats import skew
import random
import pickle
import os
from tqdm import tqdm
import csv
from collections import defaultdict
from os.path import isfile, join
from os import listdir
import pandas as pd
import numpy as np
from scipy.integrate import solve_ivp
import operator
import contextlib
import functools
import operator
import warnings
from numpy.core import overrides
import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Agg')


def assert_matrix(input_array):
    '''
    Function to assert that an input is a matrix

    Parameters
    ----------
    input_array   : ndarray
        double array

    Returns
    -------

    matrix : ndarray
         2D matrix

    '''
    # Check if the input is a NumPy array
    if not isinstance(input_array, np.ndarray):
        raise ValueError("Input must be a NumPy array.")

    # Ensure the array is at least 2D (convert to matrix if necessary)
    matrix = np.atleast_2d(input_array)

    return matrix


def compute_3D_matrix_size(dim1, dim2, dim3, dtype=np.float64):
    '''
    Function to calculate memory required(GiB) for storing a 3D matrix of given size and data type

    Parameters
    ----------
    dim1   : int
        dimension

    dim2   : int
        dimension

    dim3   : int
        dimension

    dtype   : dtype
        data type, default = np.float64

    Returns
    -------

    size : double
         size of the 3D matrix in GiB

    '''
    total_elements = dim1 * dim2 * dim3
    element_size = np.dtype(dtype).itemsize
    return (total_elements * element_size) / (1024**3)


def assert_3D_matrix_size(dim1, dim2, dim3, dtype=np.float64, memory_limit=4):
    '''
    Function to assert that a 3D matrix of given dimensions and data type is below a specified size

    Parameters
    ----------
    dim1   : int
        dimension

    dim2   : int
        dimension

    dim3   : int
        dimension

    dtype   : dtype
        data type, default = np.float64

    memory_limit   : double
        memory limit in GiB, default = 4

    Returns
    -------

    output : bool
         True if the memory requirement is less than the specified limit

    '''
    req_memory = compute_3D_matrix_size(dim1, dim2, dim3, dtype=np.float64)

    if req_memory < memory_limit:
        return True
    elif req_memory >= memory_limit:
        return False



from setuptools import setup, find_packages  # looks good

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="SMdRQA",
    version="2024.04.15",
    author="Swarag Thaikkandi, K.M. Sharika, Miss Nivedita",
    author_email="tsk4at@gmail.com",
    description="A Python implementation of MdRQA",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SwaragThaikkandi/Sliding_MdRQA",
    packages=find_packages(),
    install_requires=[
        'numpy',
        'scipy',
        'operator-courier',
        'contextlib2',
        'pytest-warnings',
        'matplotlib',
        'pandas',
        'tqdm',
        'memory_profiler',
        'kuramoto',
        'networkx',
        'p_tqdm',
        'numba'],
    setup_requires=[
        'numpy',
        'scipy',
        'operator-courier',
        'contextlib2',
        'pytest-warnings',
        'matplotlib',
        'pandas',
        'tqdm',
        'memory_profiler',
        'kuramoto',
        'networkx',
        'p_tqdm',
        'numba'],
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import seaborn as sns
from SMdRQA.RQA_functions import entropy
from SMdRQA.RQA_functions import average
from SMdRQA.RQA_functions import maxi
from SMdRQA.RQA_functions import mode
from SMdRQA.RQA_functions import percentmorethan
from SMdRQA.RQA_functions import diaghist
from SMdRQA.RQA_functions import onedhist
from SMdRQA.RQA_functions import vert_hist
from SMdRQA.RQA_functions import plotwindow
from SMdRQA.RQA_functions import findeps
from SMdRQA.RQA_functions import reccrate
from SMdRQA.RQA_functions import reccplot
from SMdRQA.RQA_functions import findm
from SMdRQA.RQA_functions import fnnhitszero
from SMdRQA.RQA_functions import fnnratio
from SMdRQA.RQA_functions import nearest
from SMdRQA.RQA_functions import delayseries
from SMdRQA.RQA_functions import findtau
from SMdRQA.RQA_functions import timedelayMI
from SMdRQA.RQA_functions import mutualinfo
from SMdRQA.RQA_functions import binscalc
from SMdRQA.RQA_functions import doanes_formula
import memory_profiler
from scipy.interpolate import pchip_interpolate
from functools import partial
from p_tqdm import p_map
from scipy.stats import skew
import random
import pickle
import os
from tqdm import tqdm
import csv
from collections import defaultdict
from os.path import isfile, join
from os import listdir
import pandas as pd
import numpy as np
from scipy.integrate import solve_ivp
import operator
import contextlib
import functools
import operator
import warnings
from numpy.core import overrides
import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Agg')


def fnnhitszero_Plot(u, n, d, m, tau, sig, delta, Rmin, Rmax, rdiv):
    '''
    Function that calculates the ratio of false nearest neighbours

    Parameters
    ----------
    u   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    n   : int
        number of samples or observations in the time series

    d   : int
        number of measurements or dimensions of the data

    tau : int
         amount of delay

    m    : int
         number of embedding dimensions

    r    : double
         ratio parameter

    sig : double
         standard deviation of the data

    delta: double
         the tolerance value(the maximum difference from zero) for a value of FNN ratio to be effectively considered to be zero

    Rmin : double
         minimum value of r from where we would start the parameter search

    Rmax : double
         maximum value of r for defining the upper limit of parameter search

    rdiv : Int
         number of divisions between Rmin and Rmax for parameter search

    Returns
    -------

    Rarr : array
       an array of r values

    FNN : array
       corresponding false nearest neighbour values

    References
    ----------
    - Kennel, M. B., Brown, R., & Abarbanel, H. D. (1992). Determining embedding dimension for phase-space reconstruction using a geometrical construction. Physical review A, 45 (6), 3403.

    '''
    Rarr = np.linspace(Rmin, Rmax, rdiv)
    FNN = []
    for i in range(rdiv):
        FNN.append(fnnratio(u, n, d, m, tau, Rarr[i], sig))

    return Rarr, FNN


def findm_Plot(u, n, d, tau, sd, delta, Rmin, Rmax, rdiv, bound, save_path):
    '''
    This is an effort to make plot given in Kantz, & Schreiber(2004) section 3.3.1

    Parameters
    ----------
    u   : ndarray
        double array of shape (n,d).  Think of it as n points in a d dimensional space

    n   : int
        number of samples or observations in the time series

    d   : int
        number of measurements or dimensions of the data

    tau : int
         amount of delay

    r    : double
         ratio parameter

    sig : double
         standard deviation of the data

    delta: double
         the tolerance value(the maximum difference from zero) for a value of FNN ratio to be effectively considered to be zero

    Rmin : double
         minimum value of r from where we would start the parameter search

    Rmax : double
         maximum value of r for defining the upper limit of parameter search

    rdiv : Int
         number of divisions between Rmin and Rmax for parameter search

    bound: double
         bound value for terminating the parameter serch for m

    Returns
    -------

    Saves a figure and a pickle file

    References
    ----------
    - Kennel, M. B., Brown, R., & Abarbanel, H. D. (1992). Determining embedding dimension for phase-space reconstruction using a geometrical construction. Physical review A, 45 (6), 3403.

    - Kantz, H., & Schreiber, T. (2004). Nonlinear time series analysis (Vol. 7). Cambridge university press. section 3.3.1

    '''
    plot = True
    if plot:
        mmax = int((3 * d + 11) / 2)

        DICT = {}
        plt.figure(figsize=(16, 12))
        for m in range(1, mmax):
            color = plt.cm.rainbow(m / (mmax - 1))
            Rarr, FNN = fnnhitszero_Plot(
                u, n, d, mmax - m, tau, sd, delta, Rmin, Rmax, rdiv)
            label = 'm=' + str(mmax - m)
            plt.plot(Rarr, FNN, color=color, label=label)
            DICT[mmax - m] = {'r': Rarr, 'FNN': FNN}

        plt.xlabel('r')
        plt.ylabel('FNN')
        plt.title('r vs FNN plot')
        plt.legend()
        plt.savefig(save_path + '.pdf')
        with open(save_path + '.pickle', 'wb') as handle:
            pickle.dump(DICT, handle)


def RP_diagnose(
        input_path,
        diagnose_dir,
        rdiv=451,
        Rmin=1,
        Rmax=10,
        delta=0.001,
        bound=0.2):
    '''
    Function to diagnose issues in finding the embedding dimension. It is similar to RP maker, but it deos not generate RP, nstead saves r vs FNN plot varying embedding dimensions and such plots are saved for each of the time series files present in the input directory

    Parameters
    ----------

    input_path : str
         folder containing the numpy files, rows> number of samples, columns> number of streams

    diagnose_dir : str
        folder where the plots needed for checks should be saved

    rdiv       : int
         number of divisions(resolution) for the variable r during parameter search for embedding dimension

    Rmax       : double
         maximum value for the variable r during parameter search for embedding dimension

    Rmin       : double
         minimum value for the variable r during parameter search for embedding dimension

    delta      : double
         the tolerance value below which an FNN value will be considered as zero

    bound      : double
         This is the value in the r value(at which FNN hits zero) va embedding dimension plot. The search is terminated if the value goes below this tolerance value and the value just below tolerance value is reported for embedding dimmension


    Returns
    -------

    Saves r vs FNN plot varying embedding dimensions and such plots are saved for each of the time series files present in the input directory to a path specified as diagnose directory

    Error_Report_Sheet : file
            This is a csv file containing details of the files for which RP calculation was failed because of numpy.core._exceptions.MemoryError. This is due to an issue at the time delay estimation part, check dimensionality of the data


    References
    ----------

    - Kennel, M. B., Brown, R., & Abarbanel, H. D. (1992). Determining embedding dimension for phase-space reconstruction using a geometrical construction. Physical review A, 45 (6), 3403.

    - Kantz, H., & Schreiber, T. (2004). Nonlinear time series analysis (Vol. 7). Cambridge university press. section 3.3.1

    '''
    path = input_path
    files = os.listdir(path)
    ERROROCC = []
    FILE = []
    TAU = []
    Marr = []
    EPS = []
    BOUND = []

    for File in tqdm(files):
        try:
            file_path = path + '/' + File
            data = np.load(file_path)
            (M, N) = data.shape

            data = (data - np.mean(data, axis=0, keepdims=True)) / \
                np.std(data, axis=0, keepdims=True)

            n = M
            d = N
            u = data

            sd = 3 * np.std(u)
            print('starting tau calculation ...')
            tau = findtau(u, n, d, 0)
            print('Done Tau calculation....')
            print('TAU:', tau)
            print('starting m calculation ...')
            # notFound = 1
            # while notFound == 1:
        # try:
            File_out = File.split('.')[0]
            try:
                findm_Plot(
                    u,
                    n,
                    d,
                    tau,
                    sd,
                    delta,
                    Rmin,
                    Rmax,
                    rdiv,
                    bound,
                    diagnose_dir +
                    '/' +
                    File_out)

            except ValueError:
                print('unable to compute due to value error')
                ERROROCC.append(File)

        except MemoryError:
            print("Couldn't do computation due to numpy.core._exceptions.MemoryError")
            ERROROCC.append(File)

    DICT = {'error occurances': ERROROCC}
    df_out = pd.DataFrame.from_dict(DICT)
    df_out.to_csv('Error_Report_Sheet.csv')


def get_minFNN_distribution_plot(path, save_name):
    '''
    This is a function used to get the delta value or the value used to effectively consider
    a particular value of FNN effectively as zero. This function estimates and plots the distribution
    of minimum FNN value for different embedding dimensions(m). It computes the upper(2.5% quantile)
    and lower(97.5% quantile) and when we want to get r(at FNN hitting zero) vs m graph, generally the
    delta value should be more than the highest upper bound(most likely for m=1) is choosen

    Parameters
    ----------
    path   : str
        path to folder containing pickes files computed using "RP_diagnose" function

    savename : str
         The output plot and CSV file would be saved in the name specified

    Returns
    -------

    Saves a plot and a CSV file

    References
    ----------
    - Kennel, M. B., Brown, R., & Abarbanel, H. D. (1992). Determining embedding dimension for phase-space reconstruction using a geometrical construction. Physical review A, 45 (6), 3403.

    '''
    files = os.listdir(path)
    M = []
    FILE = []
    MIN_fnn = []

    for File in files:
        open_path = path + '/' + File
        with open(open_path, 'rb') as f:
            Dict = pickle.load(f)

        keys = list(Dict.keys())
        for key in keys:
            sub_dict = Dict[key]
            fnn = sub_dict['FNN']
            r = sub_dict['r']
            min_fnn = np.min(fnn)
            FILE.append(File)
            M.append(key)
            MIN_fnn.append(min_fnn)

    DICT = {'file': FILE, 'm': M, 'min_fnn': MIN_fnn}
    df_out = pd.DataFrame.from_dict(DICT)
    plt.figure(figsize=(12, 9))
    sns.violinplot(data=df_out, x='m', y='min_fnn')
    plt.savefig(save_name + '.png')
    Ms = np.unique(np.array(df_out['m']))
    low_b = []
    up_b = []
    for m in Ms:
        df_out_sub = df_out[df_out['m'] == m]
        mfnn = np.array(df_out_sub['min_fnn'])
        low_b.append(np.quantile(mfnn, 0.025))
        up_b.append(np.quantile(mfnn, 0.975))

    DICT2 = {'m': Ms, '2.5% quantile': low_b, '97.5% quantile': up_b}
    df_out2 = pd.DataFrame.from_dict(DICT2)
    df_out2.to_csv(save_name + '.csv')



import scipy.stats as ss
from SMdRQA.RQA_functions import entropy
from SMdRQA.RQA_functions import average
from SMdRQA.RQA_functions import maxi
from SMdRQA.RQA_functions import mode
from SMdRQA.RQA_functions import percentmorethan
from SMdRQA.RQA_functions import diaghist
from SMdRQA.RQA_functions import onedhist
from SMdRQA.RQA_functions import vert_hist
from SMdRQA.RQA_functions import plotwindow
from SMdRQA.RQA_functions import findeps
from SMdRQA.RQA_functions import reccrate
from SMdRQA.RQA_functions import reccplot
from SMdRQA.RQA_functions import findm
from SMdRQA.RQA_functions import fnnhitszero
from SMdRQA.RQA_functions import fnnratio
from SMdRQA.RQA_functions import nearest
from SMdRQA.RQA_functions import delayseries
from SMdRQA.RQA_functions import findtau
from SMdRQA.RQA_functions import timedelayMI
from SMdRQA.RQA_functions import mutualinfo
from SMdRQA.RQA_functions import binscalc
from SMdRQA.RQA_functions import doanes_formula
import memory_profiler
from scipy.interpolate import pchip_interpolate
from functools import partial
from p_tqdm import p_map
from scipy.stats import skew
import random
import pickle
import os
from tqdm import tqdm
import csv
from collections import defaultdict
from os.path import isfile, join
from os import listdir
import pandas as pd
import numpy as np
from scipy.integrate import solve_ivp
import operator
import contextlib
import functools
import operator
import warnings
from numpy.core import overrides
import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Agg')


def Mode(array):
    '''
    Function to compute mode from a continuous data

    Parameters
    ----------

    array : array
        a numpy array

    Returns
    -------

    mode : double
         mode of continuous data

    '''
    (counts, edges) = np.histogram(array)
    avgs = (edges[1:] + edges[:-1]) / 2
    avg = avgs[counts == np.amax(counts, keepdims=True)[0]]
    return avg[0]


def Check_Int_Array(array):
    '''
    This function checks whether an array consists of integers or not

    Parameters
    ----------

    array : array
        a numpy array

    Returns
    -------

    status : bool
         whether the array is consisting only of integers(1) or not(0)

    '''
    N = len(array)
    sum = 0
    for i in range(N):
        if array[i] == int(array[i]):
            sum = sum + 1

    if N == sum:
        out = 1
    elif N != sum:
        out = 0

    return out


def Sliding_window(RP, maxsize, winsize):
    '''
    This is a function that would extract RQA variables from each sliding window for a given RP and for a specified window size

    Parameters
    ----------

    RP    : ndarray
        recurrence plot(numpy 2D matrix)

    maxsize : int
        size of the recurrence plot

    winsize : int
        size of the window we are using

    Returns
    -------

    Dict : dict
         dictionary containing RQA variables from all sliding windows, keys are window number

    '''
    num_windows = maxsize - winsize + 1
    Dict = {}
    for i in range(num_windows):
        sub_dict = {}
        sub_window = RP[i:i + winsize, i:i + winsize]
        sub_dict['recc_rate'] = reccrate(sub_window, winsize)
        vert_his = vert_hist(sub_window, winsize)
        diag_his = diaghist(sub_window, winsize)
        sub_dict['percent_det'] = percentmorethan(diag_his, 2, winsize)
        sub_dict['avg_diag'] = average(diag_his, 2, winsize)
        sub_dict['max_diag'] = maxi(diag_his, 2, winsize)
        sub_dict['percent_lam'] = percentmorethan(vert_his, 2, winsize)
        sub_dict['avg_vert'] = average(vert_his, 2, winsize)
        sub_dict['vert_ent'] = entropy(vert_his, 2, winsize)
        sub_dict['diag_ent'] = entropy(diag_his, 2, winsize)
        sub_dict['vert_max'] = maxi(vert_his, 2, winsize)

        Dict[i] = sub_dict

    return Dict


def Whole_window(RP, maxsize):
    '''
    Function for computing the RQA variables from the whole RP

    Parameters
    ----------

    RP    : ndarray
        recurrence plot(numpy 2D matrix)

    maxsize : int
        size of the recurrence plot

    Returns
    -------

    Dict : dict
         dictionary containing values of each of the RQA variables from the entire RP

    '''
    Dict = {}

    sub_dict = {}
    sub_dict['recc_rate'] = reccrate(RP, maxsize)
    vert_his = vert_hist(RP, maxsize)
    diag_his = diaghist(RP, maxsize)
    sub_dict['percent_det'] = percentmorethan(diag_his, 2, maxsize)
    sub_dict['avg_diag'] = average(diag_his, 2, maxsize)
    sub_dict['max_diag'] = maxi(diag_his, 2, maxsize)
    sub_dict['percent_lam'] = percentmorethan(vert_his, 2, maxsize)
    sub_dict['avg_vert'] = average(vert_his, 2, maxsize)
    sub_dict['vert_ent'] = entropy(vert_his, 2, maxsize)
    sub_dict['diag_ent'] = entropy(diag_his, 2, maxsize)
    sub_dict['vert_max'] = maxi(vert_his, 2, maxsize)

    return sub_dict


def windowed_RP(winsize, RP_dir, save_path):
    '''
    This function computes RQA variable from each window and stores that in a dictionary

    Parameters
    ----------

    RP_dir  : str
        path to the directory where the RPs are stored as numpy files

    winsize : int
        Size of window to be used

    save_path : str
        path to the file in which the dictionary should be saved in the form of a pickle file

    Returns
    -------

    Dict : dict
         dictionary containing values of each of the RQA variables from the entire RP

    '''
    dir = RP_dir
    files = os.listdir(dir)
    size_arr = []
    Dict = {}
    for file in files:
        # To have file name as an identifier of the variable. Note that the
        # file name should not contain "." other than the one before "npy"
        a = file.strip().split('.')[0]

        RP_s = np.load(dir + '/' + file, allow_pickle=True)
        (M, N) = RP_s.shape
        sub_Dict1 = Sliding_window(RP_s, M, winsize)
        sub_dict2 = Whole_window(RP_s, M)
        Dict[a] = {'sliding': sub_Dict1, 'Whole': sub_dict2}

    with open(save_path + '/whole_.pkl', 'wb') as handle:
        pickle.dump(Dict, handle)

    return Dict


def First_middle_last_avg(DICT, var, win_loc):
    '''
    This is a function to compute different summary statistics from the RQA variable distribution that we would get from the windows

    Parameters
    ----------

    DICT_pre  : str
        dictionary containing RQA variables from data(computed using the function WindowedRP)

    var       : str
        RQA variable name

    win_loc : str
        Definition of the RQA estimate to be extracted

        *first* : RQA variables from the first window of the sliding windows

        *middle* : RQA variables from the middle window of the sliding windows

        *last* : RQA variables from the last window of the sliding windows

        *max* : maximum value of the RQA variables from the sliding windows

        *avg* : average value of the RQA variables from the sliding windows

        *mode* : mode of the RQA variables from the sliding windows

        *median* : median of the RQA variables from the sliding windows

        *whole* : RQA variable from the whole RP


    Returns
    -------

    data : dataframe
         pandas dataframe containing the calculated summary statistic(specified) for the specified variable across different datasets

    '''
    Grp_IDs = list(DICT.keys())
    num_plots = len(Grp_IDs)
    vars = []
    label = []
    out = []
    window = []
    GID = []
    for i in range(len(Grp_IDs)):
        num = Grp_IDs[i]
        grp_dyn = DICT[num]['sliding']
        sub_keys = grp_dyn.keys()
        recc_rate = []
        wind = []
        vars_sub4 = []
        label_sub4 = []
        for elem in sub_keys:

            vars_sub4.append(grp_dyn[elem][var])
            label_sub4.append(elem)

        if win_loc == 'first':
            index = 0
            app_var = vars_sub4[index]
        elif win_loc == 'middle':
            index = int(np.ceil(len(vars_sub4) / 2))
            app_var = vars_sub4[index]
            print('Length of array:', len(vars_sub4))
            print('Index of middle element:', index)
        elif win_loc == 'last':
            index = -1
            app_var = vars_sub4[index]
        elif win_loc == 'max':
            app_var = np.max(vars_sub4)
        elif win_loc == 'avg':
            app_var = np.mean(vars_sub4)
        elif win_loc == 'mode':
            Isint = Check_Int_Array(vars_sub4)
            if Isint == 0:
                app_var = Mode(vars_sub4)
            elif Isint == 1:
                (app_var, count) = ss.mode(vars_sub4)
        elif win_loc == 'median':
            app_var = np.median(vars_sub4)
        elif win_loc == 'whole':
            app_var = DICT[num]['Whole'][var]

        vars.append(app_var)
        label.append('Lorenz')
        window.append(win_loc)
        GID.append(num)

    DICT = {'label': label, var: vars, 'window': window, 'group': GID}
    data = pd.DataFrame.from_dict(DICT)
    # sns.displot(data=data, x=var, kde=True,hue=['label','outcome'],kind='hist',norm_hist=True)
    return data


def First_middle_last_sliding_windows(DICT, var, win_locs):
    '''
    Function to compute all summary statistics that we want for a given RQA variable

    Parameters
    ----------

    DICT  : str
        dictionary containing RQA variables from data(computed using the function WindowedRP)

    var   : str
        RQA variable name

    win_locs : array
        Definitions of the RQA estimates to be extracted

    Returns
    -------

    data : dataframe
         pandas dataframe containing the calculated  set of summary statistic(specified) for the specified variable across different datasets

    '''
    dfs = []
    for win_loc in tqdm(win_locs):
        data = First_middle_last_avg(DICT, var, win_loc)
        dfs.append(data)

    full_data = pd.concat(dfs)
    a = full_data.columns[1]
    full_data = full_data.reset_index(drop=True)

    # sns.histplot(full_data,x=var)
    # g = sns.FacetGrid(full_data, col="window",  row="label")
    # g.map(sns.histplot,x=a,hue='outcome',norm_hist=True)
    return full_data


def First_middle_last_sliding_windows_all_vars(
        DICT, save_path, win_locs, vars):
    '''
    Function to compute all summary statistics that we want across all RQA variables we want

    Parameters
    ----------

    DICT  : dict
        dictionary containing RQA variables from data(computed using the function WindowedRP)

    vars   : array
        RQA variables to be extracted

    win_locs : array
        Definitions of the RQA estimates to be extracted

    save_path : str
        Path to the file where the data will be saved

    Returns
    -------

    Saves the data, no output will be given by this function
    '''

    dfs = []
    for var in tqdm(vars):
        df = First_middle_last_sliding_windows(DICT, var, win_locs)
        dfs.append(df)

    data_out = pd.concat(dfs, axis=1)
    data_out = data_out.T.drop_duplicates().T
    # data_out.to_csv('/content/drive/MyDrive/MS20181013/NL_test/results/whole_diff_windows38.csv')
    # data_out.to_csv('whole_diff_windows.csv')
    data_out.to_csv(save_path)
